"""
Functions for retrieving information on parameters, typically provided via
both str and :class:`.ParameterType` arguments.
"""

from collections import defaultdict
from typing import Union as _Union

from . import ParameterType, ParameterCategory
from ._maps import (
    NAME_PTYPE_MAP as _NAME_PTYPE_MAP,
    PTYPE_NAME_MAP as _PTYPE_NAME_MAP,
    PTYPE_DESCR_MAP as _PTYPE_DESCR_MAP,
    PTYPE_CATEGORY_MAP as _PTYPE_CAT_MAP,
)

from ..errors import CavCalcError


def get_default_units(ptype: _Union[str, ParameterType]):
    """Get the default units, as a string, of the given parameter type.

    The default units value is obtained from the loaded config file(s); i.e.
    the first instance of the corresponding option in the load order: current
    working directory -> user config directory -> package install location.

    Parameters
    ----------
    ptype : str | :class:`.ParameterType`
        The name (as it appears in the config file for example), or :class:`.ParameterType`,
        of the parameter.

    Returns
    -------
    units : Optional[str]
        A string representing the units for the given parameter, or ``None``, if ``ptype``
        does not correspond to any valid option in any of the loaded config files.
    """
    from .. import _CONFIG

    if isinstance(ptype, ParameterType):
        ptype = _PTYPE_NAME_MAP[ptype]

    return _CONFIG["units"].get(ptype)


def get_name(ptype: ParameterType):
    """Obtain the name of a parameter from the :class:`.ParameterType` literal.

    Parameters
    ----------
    ptype : :class:`.ParameterType`
        The parameter type.

    Returns
    -------
    name : Optional[str]
        The name of the parameter, or ``None`` if ``ptype`` does not correspond to
        any parameter.
    """
    return _PTYPE_NAME_MAP.get(ptype)


def get_names(*args) -> tuple[str]:
    """Retrieve a tuple of the names of all the parameters exposed by cavcalc, or
    all the parameters in the categories given in ``args``.

    Parameters
    ----------
    args : Sequence[:class:`.ParameterCategory`]
        A sequence of :class:`.ParameterCategory` literals from which
        to obtain parameter names. If none given, then all parameter
        names will be returned.

    Returns
    -------
    names : tuple[str]
        The names of the parameters.
    """
    if not args:
        return tuple(_NAME_PTYPE_MAP.keys())

    if not all(isinstance(arg, ParameterCategory) for arg in args):
        raise CavCalcError("All args must be of type ParameterCategory.")

    cat_to_types = defaultdict(set)
    for ptype, cat in _PTYPE_CAT_MAP.items():
        cat_to_types[cat].add(ptype)

    ptypes = set()
    for arg in args:
        ptypes.update(cat_to_types[arg])

    return tuple(get_name(ptype) for ptype in ptypes)


def get_names_descriptions() -> dict[str, str]:
    """Get a dictionary of all the parameter names with a description of each of these.

    Returns
    -------
    details : dict[str, str]
        The names of all the parameters (keys), with their descriptions (values).
    """
    return {get_name(ptype): descr for ptype, descr in _PTYPE_DESCR_MAP.items()}


def get_type(name: str):
    """Get the :class:`.ParameterType` literal corresponding to the
    given name of the parameter.

    Parameters
    ----------
    name : str
        The name of the parameter.

    Returns
    -------
    ptype : Optional[:class:`.ParameterType`]
        The parameter type of the parameter, or ``None`` if ``name`` does not
        correspond to any parameter.
    """
    return _NAME_PTYPE_MAP.get(name)


# TODO (sjr) Get rid of below eventually, as should be able to specify any
#            parameter as an arg or target


def get_valid_arguments():
    """Get the names of all the parameters which can be used as arguments
    in a :func:`.calculate` / CLI call.

    Returns
    -------
    names : tuple[str]
        The valid argument parameter names.
    """
    non_arg_ptypes = (
        ParameterType.FINESSE,
        ParameterType.FSR,
        ParameterType.FWHM,
        ParameterType.MODESEP,
        ParameterType.POLE,
        ParameterType.RES_ENHANCE_INTERNAL,
        ParameterType.RES_ENHANCE_EXTERNAL,
        ParameterType.TRANSMISSION_INTENSITY_FRAC,
        ParameterType.WAISTPOS,
        ParameterType.WAISTSIZE,
    )
    return tuple(name for name in get_names() if get_type(name) not in non_arg_ptypes)


def get_valid_targets():
    """Get the names of all the parameters which can be specified as targets
    in a :func:`.calculate` / CLI call.

    Returns
    -------
    names : tuple[str]
        The valid target parameter names.
    """
    non_tgt_ptypes = (
        ParameterType.CAV_LENGTH,
        ParameterType.LOSS_M1,
        ParameterType.LOSS_M2,
    )
    return tuple(name for name in get_names() if get_type(name) not in non_tgt_ptypes)



import argparse
from enum import Enum

from ._arguments._groups import make_arguments
from ._arguments._container import _Arguments
from .parameters import valid_arguments


class SessionType(Enum):
    CLI = 0
    PyAPI = 1
    # TODO (sjr) Maybe add a JupyterAPI literal here in the future
    #            if we want different behaviour for plotting / printing
    #            when in a Jupyter environment


class _CavCalcSession:
    def __init__(self):
        # Default to Python API mode, and only reset this to CLI if we're in __main__:main
        self.mode = SessionType.PyAPI
        self.parser = argparse.ArgumentParser(
            description="A tool to calculate Fabry-Perot cavity parameters."
        )
        make_arguments(self.parser)

    @property
    def is_api_mode(self) -> bool:
        return self.mode == SessionType.PyAPI

    @property
    def is_cli_mode(self) -> bool:
        return self.mode == SessionType.CLI

    def parse_args(self, command: list[str] = None):
        # Get the order in which each physical argument was specified
        opts = filter(lambda x: x.startswith("-"), command)
        phys_args = tuple(arg for opt in opts if (arg := opt.split("-")[1]) in valid_arguments)
        arg_order = {arg_name: i for i, arg_name in enumerate(phys_args)}

        args = _Arguments()
        self.parser.parse_args(command, namespace=args)
        args.process(arg_order)
        return args



"""
The output objects which store the results from any call to :func:`.calculate`.
"""

import abc
import pickle
from typing import Sequence as _Sequence, Union as _Union

from . import Q_
from .errors import CavCalcError
from ._exiters import (
    bug as _bug,
    quit_print as _quit_print,
)
from .parameters import (
    ArgParameter,
    TargetParameter,
    ParameterCategory,
    ParameterType,
)


def load(file: str):
    """Load a :class:`.SingleOutput` or :class:`.MultiOutput` instance
    from a "pickled" binary file.

    Parameters
    ----------
    file : str
        The name (and path if it isn't in the current working directory)
        of the file to open.

    Returns
    -------
    output : :class:`.SingleOutput` | :class:`.MultiOutput`
        The cavcalc output instance.

    Raises
    ------
    cce : :class:`.CavCalcError`
        If the type of the loaded object is not :class:`.SingleOutput`
        or :class:`.MultiOutput`.
    """
    with open(file, "rb") as f:
        output = pickle.load(f)
        if not isinstance(output, (SingleOutput, MultiOutput)):
            raise CavCalcError(f"Unrecognised serialized object in {file}")

    return output


def _param_str_gen(params: _Sequence[_Union[ArgParameter, TargetParameter]]):
    return (f"{'' if getattr(p, 'was_used', True) else '[UNUSED] '}{p}" for p in params)


class BaseOutput(abc.ABC):
    """Abstract base class for output objects. Both :class:`.SingleOutput` and
    :class:`.MultiOutput` derive from this type."""

    category_order = (
        ParameterCategory.Frequency,
        ParameterCategory.Power,
        ParameterCategory.Distance,
        ParameterCategory.Curvature,
        ParameterCategory.BeamRadius,
        ParameterCategory.Stability,
        ParameterCategory.Phase,
        ParameterCategory.Angle,
        ParameterCategory.Wave,
    )

    def __init__(self, phys_args: tuple[ArgParameter]):
        self.__given = phys_args

    def __str__(self) -> str:
        s = "Given:\n\t"
        for category in self.category_order:
            if args := tuple(filter(lambda p: p.category == category, self.__given)):
                s += "\n\t".join(_param_str_gen(sorted(args, key=lambda a: a.ptype.value)))
                s += "\n\n\t"

        return s.strip() + "\n\n"

    @property
    def given(self) -> dict[str, ArgParameter]:
        """A dictionary of the given parameter names with their corresponding
        :class:`.ArgParameter` values."""
        return {arg_p.name: arg_p for arg_p in self.__given}

    def print(self, **kwargs):
        """Prints a string representation of the output, in an identical style to the CLI.

        Parameters
        ----------
        kwargs : Keyword Arguments
            Optional arguments to pass to Python ``print`` function.
        """
        print(str(self), **kwargs)

    def save(self, file: str):
        """Saves the output object to the specified ``file``, in a binary
        serialised Python object format (via ``pickle``).

        Parameters
        ----------
        file : str
            The file name (and path if saving to a location outside of the current
            working directory) to save this output instance to.
        """
        with open(file, "wb") as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)


class SingleOutput(BaseOutput):
    """Storage and manipulation of a single target parameter computed via :func:`.calculate`."""

    def __init__(self, target: TargetParameter, phys_args: tuple[ArgParameter]):
        super().__init__(phys_args)
        self.__target = target

    def __str__(self) -> str:
        return super().__str__() + "Computed:\n\t" + "".join(_param_str_gen((self.result,)))

    @property
    def name(self):
        """A short-hand for ``out.result.name``."""
        return self.__target.name

    @property
    def result(self):
        """The result itself as a :class:`.TargetParameter` instance.

        .. hint::

            Use ``result.value`` to get the ``pint.Quantity`` object, ``result.value.m``
            to obtain the numerical value, and ``result.value.u`` to retrieve the units
            of this result.
        """
        return self.__target

    def convert(self, units: str):
        """Attempts to convert the result quantity, in-place, to the
        specified units.

        .. warning::

            This method converts the actual quantity held by this output. If you
            do not want to modify this object, then instead do::

                q = out.result.to(units)

            to get a new quantity object, ``q``, in the given units.

        Parameters
        ----------
        units : str
            The units to convert to, as a string or ``cavcalc.ureg.<units>`` object.
        """
        self.__target = TargetParameter(self.result.name, self.result.value.to(units))

    def plot(
        self,
        xlim: _Union[str, _Sequence[float]] = "auto",
        ylim: _Union[str, _Sequence[float]] = "auto",
        zlim: _Union[str, _Sequence[float]] = "auto",
        filename: str = None,
        logx: bool = False,
        logy: bool = False,
        logz: bool = False,
        cmap: str = None,
        show: bool = True,
        fig=None,
        **kwargs,
    ):
        """Plots the result against the given array-like argument dependencies (if any).

        If only one of the given arguments to the :func:`.calculate` call was array-like,
        then this will plot a standard line plot. Otherwise an image plot, via
        :func:`matplotlib.pyplot.imshow`, will be produced.

        Parameters
        ----------
        xlim : str | Iterable[float], optional; default = "auto"
            The limits of the x-axis. Defaults to ``"auto"`` such that matplotlib
            determines these automatically. This can be set to ``"data"`` so that
            the exact limits of the x-axis data are used. An iterable of length
            two can be used to manually override the x-axis limits; these will
            simply be forwarded to :func:`matplotlib.pyplot.xlim`.

        ylim : str | Iterable[float], optional; default = "auto"
            The limits of the y-axis. Defaults to ``"auto"`` such that matplotlib
            determines these automatically. This can be set to ``"data"`` so that
            the exact limits of the y-axis data are used. An iterable of length
            two can be used to manually override the y-axis limits; these will
            simply be forwarded to :func:`matplotlib.pyplot.ylim`.

        zlim : str | Iterable[float], optional; default = "auto"
            The limits of the color-bar scale for image plots. Ignored if
            this is a line plot.

        filename : str, optional; default = None
            A filename to save the figure to.

        logx : bool, optional; default = False
            Whether to use a log-scale on the x-axis. Ignored if this is an image plot.

        logy : bool, optional; default = False
            Whether to use a log-scale on the y-axis. Ignored if this is an image plot.

        logz : bool, optional; default = False
            Whether to use a log normalisation of the image plot Z data. Ignored if
            this is a line plot.

        cmap : str | :class:`matplotlib.colors.ListedColormap`, optional; default = None
            Name, or instance, of the colormap to use for image plots. If not given,
            the colormap will default to the relevant ``matplotlib.rcParams`` value.
            Ignored if this is a line plot.

        show : bool, optional; default = False
            Shows the resulting figure, via :func:`matplotlib.pyplot.show`, if true.

        fig : :class:`matplotlib.figure.Figure`, optional; default = None
            An optional pre-existing figure object to draw on. If not given, then a new
            figure is created and used.

        kwargs : Keyword arguments
            Optional keyword arguments to pass to the relevant matplotlib plot or imshow
            call(s).

        Returns
        -------
        fig : :class:`matplotlib.figure.Figure`
            The matplotlib Figure object.

        Raises
        ------
        cce : :class:`.CavCalcError`
            If the value of :attr:`.SingleOutput.result` is not array-like (i.e. no
            dependencies of the target parameter were array-like, when making the
            :func:`.calculate` call).
        """
        if not self.__target.is_array:
            _quit_print(
                f"Value of target '{self.__target.name}' is not array-like! Did "
                "you forget to set one of the dependent arguments to an array?"
            )

        given = self.given.values()
        xs = tuple(sorted(filter(lambda p: p.is_array, given), key=lambda p: (p.axis, p.index)))
        num_x_arrays = len(xs)
        if not num_x_arrays:
            _bug(
                f"Somehow none of the given args, ({', '.join(self.given.keys())}), "
                f"are array-like, but the target value, {self.__target.name}, is!"
            )

        plotting_dims = max(len(x.value.m.shape) for x in xs)
        if plotting_dims > 2:
            _quit_print(f"Cannot plot results over {plotting_dims} dimensions!")

        import matplotlib.pyplot as plt

        make_subplots = False
        if fig is None:
            fig = plt.figure()
        else:
            if fig.axes:
                plt.sca(fig.axes[0])
            else:
                make_subplots = True

        def _make_label(p: ArgParameter):
            return p.description + (f" [{p.value.units:~}]" if not p.is_unitless else "")

        # The non-array arguments specified
        if fixed := tuple(filter(lambda p: p.is_scalar and p.was_used, given)):
            fixed_params_str = "with: " + ", ".join(f"{p.symbol_str} = {p.value:~}" for p in fixed)
        else:
            fixed_params_str = ""

        if plotting_dims == 1 or num_x_arrays == 1:
            if num_x_arrays > 2:
                _quit_print("Plots with more than two x-axes are not supported.")

            if num_x_arrays == 1:
                (x,) = xs
            else:
                x, x2 = xs
            y = self.__target

            # This typically occurs when plotting a MultiOutput, where a target was computed
            # over a mesh-grid but only depended on one of the grid parameters. So here we
            # just grab the first row of the data and plot that; as all rows should be equal.
            # TODO (sjr) Might be better to do this more automatically in MultiOutput, via some
            #            reduce_grid method which gets applied to any target where above is true.
            if plotting_dims > 1:
                x = ArgParameter(x.name, Q_(x.value.m[0], x.value.units), axis=x.axis)
                y = TargetParameter(y.name, Q_(y.value.m[0], y.value.units))

            if make_subplots:
                plt.sca(fig.subplots())

            plot_func = plt.plot
            if logx and logy:
                plot_func = plt.loglog
            elif logx:
                plot_func = plt.semilogx
            elif logy:
                plot_func = plt.semilogy

            if num_x_arrays == 1:
                label = f"{y.symbol_str}({x.symbol_str}) " + fixed_params_str
            else:
                label = f"{y.symbol_str}({{{x.symbol_str}, {x2.symbol_str}}}) " + fixed_params_str

            try:
                plot_func(x.value.m.T, y.value.m.T, label=label, **kwargs)
            except Exception as ex:
                _quit_print(str(ex))

            if isinstance(xlim, str):
                if xlim.casefold() == "data":
                    plt.xlim(x.value.m.min(), x.value.m.max())
            else:
                if xlim:
                    plt.xlim(*xlim)
            if isinstance(ylim, str):
                if ylim.casefold() == "data":
                    plt.ylim(y.value.m.min(), y.value.m.max())
            else:
                if ylim:
                    plt.ylim(*ylim)

            plt.legend()

            if num_x_arrays > 1:
                ax1 = plt.gca()

                ax2 = plt.twiny()
                plot_func2 = ax2.plot
                if logx and logy:
                    plot_func2 = ax2.loglog
                elif logx:
                    plot_func2 = ax2.semilogx
                elif logy:
                    plot_func2 = ax2.semilogy

                # "Plot" it again over x2 to get the auto limits for this variable,
                # but don't actually draw it this time as it will be the same curve
                plot_func2(x2.value.m.T, y.value.m.T, alpha=0)
                ax2.set_xlabel(_make_label(x2))

                # Reset current state to first axes
                plt.sca(ax1)

        else:
            if num_x_arrays > 2:
                _quit_print("Image plots with multiple x-axes are not supported.")

            import matplotlib

            x, y = xs
            z = self.__target

            try:
                cmap = matplotlib.colormaps.get_cmap(cmap)
            except Exception as ex:
                _quit_print(str(ex))

            extent = [x.value.m.min(), x.value.m.max(), y.value.m.min(), y.value.m.max()]

            if logz:
                if not zlim or isinstance(zlim, str):
                    norm = plt.cm.colors.LogNorm()
                else:
                    vmin, vmax = zlim
                    if vmin <= 0:
                        vmin = None

                    norm = plt.cm.colors.LogNorm(vmin, vmax)
            else:
                if zlim and not isinstance(zlim, str):
                    norm = plt.cm.colors.Normalize(*zlim)
                else:
                    norm = None

            title = f"{z.symbol_str}({x.symbol_str}, {y.symbol_str}) " + fixed_params_str

            shape = z.value.m.shape
            if len(shape) == 2:
                if make_subplots:
                    plt.sca(fig.subplots())

                try:
                    surf = plt.imshow(
                        z.value.m.T,
                        cmap=cmap,
                        extent=extent,
                        norm=norm,
                        origin="lower",
                        aspect="auto",
                        **kwargs,
                    )
                except Exception as ex:
                    _quit_print(str(ex))

                plt.colorbar(surf, label=_make_label(z), fraction=0.046, pad=0.04)
                plt.title(title)

            elif len(shape) == 3:
                n_dual, N1, N2 = shape
                if n_dual != 2:
                    _bug(
                        f"The shape of target '{z.name}' should be (2, {N1}, {N2}) "
                        f"but got a shape of {shape} instead!"
                    )

                ax1 = fig.add_subplot(121)
                ax2 = fig.add_subplot(122)
                plt.sca(ax1)

                z1, z2 = z.value.m
                try:
                    surf1 = ax1.imshow(
                        z1.T,
                        cmap=cmap,
                        extent=extent,
                        norm=norm,
                        origin="lower",
                        aspect="auto",
                        **kwargs,
                    )
                except Exception as ex:
                    _quit_print(str(ex))

                plt.colorbar(surf1, ax=ax1, label=_make_label(z), fraction=0.09, pad=0.04)

                ax1.set_xlabel(_make_label(x))
                ax1.set_ylabel(_make_label(y))
                ax1.set_title(title)

                plt.sca(ax2)
                surf2 = ax2.imshow(
                    z2.T,
                    cmap=cmap,
                    extent=extent,
                    norm=norm,
                    origin="lower",
                    aspect="auto",
                    **kwargs,
                )

                plt.colorbar(surf2, ax=ax2, label=_make_label(z), fraction=0.09, pad=0.04)

                ax2.set_title(title)

                fig.tight_layout()

            else:
                _bug(
                    f"Dimensions, {shape}, of target '{z.name}' are incompatible "
                    "with image plotting!"
                )

        plt.xlabel(_make_label(x))
        plt.ylabel(_make_label(y))

        if filename:
            fig.savefig(filename)

        if show:
            plt.show()

        return fig


class MultiOutput(BaseOutput):
    """Storage and manipulation of multiple target parameters computed via :func:`.calculate`.

    Use :meth:`.MultiOutput.as_singles` to easily construct :class:`.SingleOutput` instances
    from a multi-output object.
    """

    def __init__(
        self, targets: dict[TargetParameter, tuple[ParameterType]], phys_args: tuple[ArgParameter]
    ):
        super().__init__(phys_args)
        self.__targets = targets

    def __getitem__(self, key: _Union[str, ParameterType, ParameterCategory]):
        if isinstance(key, str):
            try:
                return next((tgt_p for tgt_p in self.__targets if tgt_p.name == key))
            except StopIteration:
                raise CavCalcError(f"No target of name '{key}' exists in the output.")

        if isinstance(key, ParameterType):
            try:
                return next((tgt_p for tgt_p in self.__targets if tgt_p.ptype == key))
            except StopIteration:
                raise CavCalcError(f"No target of parameter type '{key}' exists in the output.")

        if isinstance(key, ParameterCategory):
            targets = tuple(filter(lambda t: t.category == key, self.__targets))
            if not targets:
                raise CavCalcError(f"No targets of parameter category '{key}' exist in the output.")

            return targets

        raise TypeError(
            f"Expected key-type of str, ParameterType or ParameterCategory, but got {type(key)}"
        )

    def __str__(self) -> str:
        s = "Computed:\n\t"
        for category in self.category_order:
            if targets := self.get(category):
                s += "\n\t".join(_param_str_gen(sorted(targets, key=lambda t: t.ptype.value)))
                s += "\n\n\t"

        return super().__str__() + s.strip()

    @property
    def results(self):
        """A dictionary of results, mapping the names of the target parameters to their values (as
        ``pint.Quantity`` instances)."""
        return {tgt_p.name: tgt_p for tgt_p in self.__targets}

    def as_single(self, ptype: _Union[str, ParameterType]) -> SingleOutput:
        """A :class:`.SingleOutput` instance constructed, from this multi-output,
        for a single desired parameter type.

        Parameters
        ----------
        ptype : str | :class:`.ParameterType`
            The parameter type.

        Returns
        -------
        single : :class:`.SingleOutput`
            The single output object for the desired parameter type.
        """
        singles = self.as_singles(ptype)
        return tuple(singles.values())[0]

    def as_singles(self, *args) -> dict[ParameterType, SingleOutput]:
        """A dictionary of :class:`.SingleOutput` instances, for given
        :class:`.ParameterType` objects, constructed from this multi-output.

        To retrieve single output objects for all target parameters in
        this multi-output, simply give no args when calling this method.

        Parameters
        ----------
        args : tuple[str | :class:`.ParameterType`]
            Parameter names or types as positional arguments. These will form the
            keys of the returned dictionary. If none are given then all target
            parameter types are used.

        Returns
        -------
        singles : dict[str | :class:`.ParameterType`, :class:`.SingleOutput`]
            The single outputs constructed for each desired parameter.
        """
        if not args:
            args = tuple(tgt_p.ptype for tgt_p in self.__targets)

        singles = {}
        for arg in args:
            if not isinstance(arg, (str, ParameterType)):
                raise CavCalcError(
                    "All arguments to as_singles must be strings or ParameterType instances."
                )

            target = next(
                (tgt_p for tgt_p in self.__targets if (tgt_p.ptype == arg or tgt_p.name == arg)),
                None,
            )
            if target is None:
                raise CavCalcError(f"No target of type '{arg}' in the output.")

            func_params = self.__targets.get(target)
            if not func_params:
                _bug(f"No function parameters for target parameter of type '{arg}'")

            phys_args = tuple(arg_p for arg_p in self.given.values() if arg_p.ptype in func_params)
            singles[arg] = SingleOutput(target, phys_args)

        return singles

    def get(self, key: _Union[str, ParameterType, ParameterCategory], default=None):
        """Retrieve a target result, or a collection of these, using a suitable key.

        If ``key`` is a string or a :class:`.ParameterType` then the corresponding target parameter
        will be returned if it exists, ``default`` is returned otherwise.

        Alternatively, ``key`` can be a :class:`.ParameterCategory`. In this case all the target
        parameters belonging to this category will be returned, or ``default`` if no targets of
        this category are present in the output.

        Parameters
        ----------
        key : str | :class:`.ParameterType` | :class:`.ParameterCategory`
            The key of the target(s) to retrieve. See above for options.

        default : Any | NoneType, optional
            The default value to return if no target(s) corresponding to ``key`` could be found.

        Returns
        -------
        target : :class:`.TargetParameter` | tuple[:class:`.TargetParameter`] | Any
            The target, or tuple of targets if ``key`` was a :class:`.ParameterCategory`, or ``default``
            if no target(s) could be obtained from the given ``key``.
        """
        try:
            return self[key]
        except CavCalcError:
            return default

    def plot(self, filename: str = None, show: bool = True, **kwargs):
        """Plots all the outputs which have array-like target parameters.

        See :meth:`.SingleOutput.plot` for plotting options, the ``**kwargs`` here
        will be forwarded to this method for each target parameter.

        Parameters
        ----------
        show : bool, optional; default = False
            Shows the resulting figure, via :func:`matplotlib.pyplot.show`, if true.

        Returns
        -------
        figures : dict[:class:`.ParameterType`, :class:`matplotlib.figure.Figure`]
            A dictionary of the parameter types mapping to their respective figure objects.
        """
        import matplotlib.pyplot as plt

        singles = self.as_singles(*(tgt_p.ptype for tgt_p in self.__targets if tgt_p.is_array))
        figures = {}
        for ptype, out in singles.items():
            if filename:
                try:
                    filename_single, ext = filename.split(".")
                except ValueError:
                    filename_single = filename
                    ext = ""

                filename_single = f"{filename_single}_{out.name}" + (f".{ext}" if ext else ext)
            else:
                filename_single = None

            fig = out.plot(filename=filename_single, show=False, **kwargs)
            figures[ptype] = fig

        if show and figures:
            plt.show()

        return figures



"""Functions to convert strings, passed as CLI arguments, to
appropriate quantities for these argument types."""

import numpy as np
import os
from pint.errors import UndefinedUnitError

from ._utils import _ReferencedQuantity
from .. import Q_, ureg
from .._exiters import quit_print
from ..parameters import valid_arguments


def _parse_data_range(*args):
    nargs = len(args)
    if nargs < 3 or nargs > 4:
        quit_print(
            f'Expected data range in format "<start> <stop> <num> [<units>]" ' f"but got: {args}"
        )

    if nargs == 3:
        start, stop, num = args
        units = ""
    else:
        start, stop, num, units = args

    for x in start, stop, num:
        xc = x.casefold()
        if "inf" in xc:
            quit_print("Encountered 'inf' in a data range. Values must be real and finite.")
        if "nan" in xc:
            quit_print("Encountered 'NaN' in a data range. Values must be real and finite.")

    try:
        start = float(start)
    except ValueError:
        quit_print(
            f"Could not convert data range start value '{start}' to a floating point number."
        )

    try:
        stop = float(stop)
    except ValueError:
        quit_print(f"Could not convert data range stop value '{stop}' to a floating point number.")

    try:
        num = int(num)
    except ValueError:
        quit_print(f"Could not convert data range num value '{num}' to an integer.")

    if num <= 0:
        quit_print(f"Number of points in data range must be a positive integer.")

    try:
        return Q_(np.linspace(start, stop, num), units)
    except UndefinedUnitError as ex:
        quit_print(str(ex))


def _parse_file(filename: str, units: str = None):
    try:
        if filename.casefold().endswith(".npy"):
            data = np.load(filename)
        else:  # load from text/csv file
            data = np.loadtxt(filename)
    except Exception as ex:
        quit_print(f"Unable to load file {filename}, error message: {ex}")

    if not np.issubdtype(data.dtype, np.number) or np.issubdtype(data.dtype, np.complex128):
        quit_print(f"Invalid data-type for parameter, expected real-valued numbers only.")

    return Q_(data, units)


def float_file_range_t(string: str):
    string = string.strip()

    # First check if it's a reference to another parameter...
    if string in valid_arguments:
        return _ReferencedQuantity(string)

    possible_file = True
    args = string.split()
    if (n_args := len(args)) > 1:
        if n_args > 2:  # ... parse range syntax into a linearly spaced array
            return _parse_data_range(*args)

        value, units = args
        try:
            ureg.Unit(units)
        except (UndefinedUnitError, ValueError) as ex:
            quit_print(f"Parsing units: {units}. " + str(ex))

        if os.path.isfile(value):
            return _parse_file(value, units)

        # if no file was specified then concatenate the value and
        # units into one string for easy conversion into Quantity
        string = value + units
        possible_file = False

    if possible_file and os.path.isfile(string):
        return _parse_file(string)

    try:
        return Q_(string)
    except UndefinedUnitError as ex:
        quit_print(str(ex))


def mesh_t(string: str):
    string = string.strip()

    if string.casefold() == "true":
        return True

    if string.casefold() == "false":
        return tuple()

    mesh_gen = (param_combo.split(",") for param_combo in string.split(";"))
    meshes = []
    params = []
    for param_combo in mesh_gen:
        if not param_combo or all(not s for s in param_combo):
            continue

        pnames = tuple(pname.strip() for pname in param_combo)
        for pname in pnames:
            if (num_occ := pnames.count(pname)) > 1:
                quit_print(
                    f"Parameter '{pname}' was given {num_occ} times in the mesh "
                    f"combination '{param_combo}'."
                )
            if pname in params:
                quit_print(
                    f"Parameter '{pname}' was already specified in a previous mesh "
                    "combination! Repeated parameter meshes are not currently supported."
                )
            params.append(pname)

        meshes.append(pnames)

    return tuple(meshes)


def limits_t(string: str):
    string = string.strip()

    if string.casefold() == "data":
        return string

    try:
        lo_s, hi_s = string.split()
    except ValueError:
        quit_print('Expected limits specified as "<low> <high>".')

    try:
        lo = None if lo_s == "None" else float(lo_s)
    except ValueError:
        quit_print(f"Lower limit: {lo_s} must be convertible to a number, or None.")

    try:
        hi = None if hi_s == "None" else float(hi_s)
    except ValueError:
        quit_print(f"Upper limit: {hi_s} must be convertible to a number, or None.")

    return lo, hi


def options_t(string: str):
    opts = string.strip().split(",")
    if any(":" not in x for x in opts):
        quit_print('Expected options argument in format "key1: value1, key2: value2"')

    return dict(map(lambda s: s.split(":"), map(lambda s: s.replace(" ", ""), opts)))



"""
A sub-package containing the Parameter classes, and some functions
for retrieving useful information from these.
"""

from .enums import ParameterType, ParameterCategory
from .parameter import ArgParameter, Parameter, TargetParameter
from .tools import get_default_units, get_valid_arguments, get_valid_targets

# TODO (sjr) Get rid of below eventually, see note in .tools

valid_arguments = get_valid_arguments()
valid_targets = get_valid_targets()



"""
Physical geometric-like parameter functions associated with non-symmetric optical cavities. Most properties
here are expressed only in terms of the invididual g-factors of the cavity mirrors, as inter-dependent
relations between these properties are computed via target chaining in the core cavcalc code.

.. note::

    Whilst these underlying functions are publicly accessible, it is recommended that you
    instead use the single-function interface :func:`.calculate` for computing any
    of the cavity properties that you need; as that will result in a nice output
    object for accessing results.
"""

import numpy as np

from .. import ureg
from ._utils import check_physical, modesep_adjust


@check_physical()
@ureg.wraps(ureg.meter, (ureg.meter, ureg.meter, "", ""))
def w1_of_g1g2(L, wl, g1, g2):
    r"""Radius of the beam, of wavelength :math:`\lambda`, at the *first* mirror of a cavity
    of length :math:`L`, with mirror g-factors of :math:`g_1` and :math:`g_2`, respectively.

    .. math::
        w_1\left(L, \lambda, g_1, g_2\right) = \sqrt{
            \frac{L \lambda}{\pi} \sqrt{\frac{g_2}{g_1(1 - g_1 g_2)}}
        }
    """
    return np.sqrt(L * wl / np.pi * np.sqrt(g2 / (g1 * (1 - g1 * g2))))


@check_physical()
@ureg.wraps(ureg.meter, (ureg.meter, ureg.meter, "", ""))
def w2_of_g1g2(L, wl, g1, g2):
    r"""Radius of the beam, of wavelength :math:`\lambda`, at the *end* mirror of a cavity
    of length :math:`L`, with mirror g-factors of :math:`g_1` and :math:`g_2`, respectively.

    .. math::
        w_2\left(L, \lambda, g_1, g_2\right) = \sqrt{
            \frac{L \lambda}{\pi} \sqrt{\frac{g_1}{g_2(1 - g_1 g_2)}}
        }
    """
    return np.sqrt(L * wl / np.pi * np.sqrt(g1 / (g2 * (1 - g1 * g2))))


@check_physical()
@ureg.wraps(ureg.meter, (ureg.meter, ureg.meter, "", ""))
def w0_of_g1g2(L, wl, g1, g2):
    r"""Radius of the beam, of wavelength :math:`\lambda`, at the waist position of a cavity
    of length :math:`L`, with mirror g-factors of :math:`g_1` and :math:`g_2`, respectively.

    .. math::
        w_0\left(L, \lambda, g_1, g_2\right) = \sqrt{
            \frac{L \lambda}{\pi}
            \sqrt{\frac{g_1 g_2 (1 - g_1 g_2)}{(g_1 + g_2 - 2 g_1 g_2)^2}}
        }
    """
    return np.sqrt(L * wl / np.pi * np.sqrt(g1 * g2 * (1 - g1 * g2) / (g1 + g2 - 2 * g1 * g2) ** 2))


@check_physical()
@ureg.wraps(ureg.meter, (ureg.meter, "", ""))
def z0_of_g1g2(L, g1, g2):
    r"""Waist position of a beam in a cavity of length :math:`L`, with mirror g-factors of
    :math:`g_1` and :math:`g_2`, respectively.

    .. math::
        z_0(L, g_1, g_2) = \frac{L g_2 (1 - g_1)}{g_1 + g_2 - 2 g_1 g_2}
    """
    return L * g2 * (1 - g1) / (g1 + g2 - 2 * g1 * g2)


def _rtgouy_of_g1g2_base(g1, g2):
    return 2 * np.arccos(0.5 * (np.sign(g1) + np.sign(g2)) * np.sqrt(g1 * g2))


@check_physical()
@ureg.wraps(ureg.radians, ("", ""))
def rtgouy_of_g1g2(g1, g2):
    r"""Round-trip gouy phase of a cavity with mirror g-factors of :math:`g_1` and
    :math:`g_2`, respectively.

    .. math::
        \psi(g_1, g_2) = 2\arccos{
            \left(
                \frac{\mathrm{sgn}\left(g_1\right) + \mathrm{sgn}\left(g_2\right)}{2}
                \sqrt{g_1 g_2}
            \right)
        }
    """
    return _rtgouy_of_g1g2_base(g1, g2)


@check_physical()
@ureg.wraps(ureg.radians, (ureg.meter, ureg.meter, "", ""))
def divang_of_g1g2(L, wl, g1, g2):
    r"""Divergence angle of the beam, of wavelength :math:`\lambda`, in a cavity of length
    :math:`L`, with mirror g-factors of :math:`g_1` and :math:`g_2`, respectively.

    .. math::
        \theta(L, \lambda, g_1, g_2) = \arctan{\left(\sqrt{
            \frac{\lambda}{L\pi} \sqrt{\frac{(g_1 + g_2 - 2 g_1 g_2)^2}{g_1 g_2 (1 - g_1 g_2)}}
        }\right)}
    """
    return np.arctan(
        np.sqrt(
            (wl / (L * np.pi)) * np.sqrt((g1 + g2 - 2 * g1 * g2) ** 2 / (g1 * g2 * (1 - g1 * g2)))
        )
    )


@check_physical()
@ureg.wraps("", (ureg.meter, ureg.meter))
def g1_of_Rc1(L, Rc1):
    r"""Stability factor of the *first* mirror of a cavity of length :math:`L`, where
    this mirror has radius of curvature :math:`R_{C,1}`.

    .. math::
        g_1(L, R_{C,1}) = 1 - \frac{L}{R_{C,1}}
    """
    return 1 - L / Rc1


@check_physical()
@ureg.wraps("", (ureg.meter, ureg.meter))
def g2_of_Rc2(L, Rc2):
    r"""Stability factor of the *end* mirror of a cavity of length :math:`L`, where
    this mirror has radius of curvature :math:`R_{C,2}`.

    .. math::
        g_2(L, R_{C,2}) = 1 - \frac{L}{R_{C,2}}
    """
    return 1 - L / Rc2


@check_physical()
@ureg.wraps("", (ureg.meter, ureg.meter, ureg.meter, ureg.meter))
def g1_of_w1w2(L, wl, w1, w2):
    r"""Stability factor of the *first* mirror of a cavity of length :math:`L`, with
    beam radii at the mirrors of :math:`w_1` and :math:`w_2`, respectively; where the
    wavelength of the beam is :math:`\lambda`.

    .. math::
        g_1(L, \lambda, w_1, w_2) = \pm \frac{w_2}{w_1} \sqrt{
            1 - \left( \frac{L\lambda}{\pi w_1 w_2} \right)^2
        }
    """
    mag = np.sqrt(1 - ((L * wl / np.pi) / (w1 * w2)) ** 2)
    return (w2 / w1) * np.array([-mag, mag])


@check_physical()
@ureg.wraps("", (ureg.meter, ureg.meter, ureg.meter, ureg.meter))
def g2_of_w1w2(L, wl, w1, w2):
    r"""Stability factor of the *end* mirror of a cavity of length :math:`L`, with
    beam radii at the mirrors of :math:`w_1` and :math:`w_2`, respectively; where the
    wavelength of the beam is :math:`\lambda`.

    .. math::
        g_1(L, \lambda, w_1, w_2) = \pm \frac{w_1}{w_2} \sqrt{
            1 - \left( \frac{L\lambda}{\pi w_1 w_2} \right)^2
        }
    """
    mag = np.sqrt(1 - ((L * wl / np.pi) / (w1 * w2)) ** 2)
    return (w1 / w2) * np.array([-mag, mag])


@check_physical()
@ureg.wraps("", ("", ""))
def g_of_g1g2(g1, g2):
    r"""Stability factor of a cavity given mirror g-factors of :math:`g_1`
    and :math:`g_2`, respectively.

    .. math::
        g(g_1, g_2) = g_1 g_2
    """
    return g1 * g2


@check_physical()
@ureg.wraps(ureg.meter, (ureg.meter, ""))
def Rc1_of_g1(L, g1):
    r"""Radius of curvature of *first* mirror of a cavity of length :math:`L`,
    with the g-factor of this mirror being :math:`g_1`.

    .. math::
        R_{C,1} = \frac{L}{1 - g_1}
    """
    return L / (1 - g1)


@check_physical()
@ureg.wraps(ureg.meter, (ureg.meter, ""))
def Rc2_of_g2(L, g2):
    r"""Radius of curvature of *end* mirror of a cavity of length :math:`L`,
    with the g-factor of this mirror being :math:`g_2`.

    .. math::
        R_{C,2} = \frac{L}{1 - g_2}
    """
    return L / (1 - g2)


@check_physical()
@ureg.wraps(ureg.hertz, (ureg.meter, "", ""))
def modesep_of_g1g2(L, g1, g2):
    r"""Mode separation frequency of a cavity of length :math:`L`, with
    mirror g-factors of :math:`g_1` and :math:`g_2`, respectively.

    In the equation below, :math:`\nu(L)` is the FSR of the cavity, and :math:`\psi(g_1, g_2)`
    is the round-trip gouy phase.

    .. math::
        \delta f(L, g_1, g_2) = \begin{cases}
            \frac{\psi(g_1, g_2)}{2\pi}\,\nu(L) \quad \text{if }\, \psi < \pi,\\
            \nu(L)\left(1 - \frac{\psi(g_1, g_2)}{2\pi}\right) \quad \text{otherwise}
        \end{cases}
    """
    rtgouy = _rtgouy_of_g1g2_base(g1, g2)
    return modesep_adjust(rtgouy, L)



import os
from pint.errors import DimensionalityError, UndefinedUnitError
import sys
import warnings

from . import _SESSION, _CONFIG
from ._exiters import quit_print
from ._handler import make_handler
from .output import load, SingleOutput


def main(args=None):
    if not args:
        args = sys.argv[1:]

    args = _SESSION.parse_args(args)

    args.plot |= args.semilogx or args.semilogy or args.loglog
    if args.loglog:
        args.semilogx = True
        args.semilogy = True

    # If any plotting needs doing, then handle style initialisations here
    if args.plot:
        if "plotting" in _CONFIG:
            import matplotlib

            plotting_conf = _CONFIG["plotting"]

            if backend := plotting_conf.get("matplotlib_backend"):
                try:
                    matplotlib.use(backend)
                except Exception as ex:
                    quit_print(str(ex))

            import matplotlib.pyplot as plt

            if style := plotting_conf.get("style"):
                if style.casefold() == "cavcalc":
                    here, _ = os.path.split(os.path.realpath(__file__))
                    cc_style_file = os.path.join(here, "_default.mplstyle")
                    if not os.path.isfile(cc_style_file):
                        warnings.warn("Could not locate cavcalc package style-sheet file!")
                    else:
                        plt.style.use(cc_style_file)
                else:
                    try:
                        plt.style.use(style)
                    except Exception as ex:
                        warnings.warn(str(ex))

    if args.loadfile:
        try:
            out = load(args.loadfile)
        except Exception as ex:
            quit_print(f"The following error occurred during load: {str(ex)}")

        if args.is_single_target:
            if isinstance(out, SingleOutput):
                if out.name != args.compute:
                    quit_print(
                        f"Mismatch between target parameter, '{args.compute}', and the "
                        f"output object parameter '{out.name}' loaded from the"
                        f"specified file '{args.loadfile}'"
                    )
            else:
                if args.compute not in out.results:
                    quit_print(
                        f"The target parameter, '{args.compute}', is not in the output"
                        f"object loaded from the specified file '{args.loadfile}'"
                    )

                out = out.as_single(args.compute)
                if args.units:
                    try:
                        out.convert(args.units)
                    except (DimensionalityError, UndefinedUnitError, ValueError) as ex:
                        quit_print(str(ex))
    else:
        out = make_handler(args).run()

    if args.savefile:
        out.save(args.savefile)

    if args.plot:
        out.plot(
            xlim=args.xlim,
            ylim=args.ylim,
            zlim=args.zlim,
            filename=args.saveplot,
            logx=args.semilogx,
            logy=args.semilogy,
            logz=args.loglog,
            cmap=args.cmap,
            show=not args.no_show_plot,
            **args.plot_opts,
        )
    else:
        print(out)

    return out



"""
A module holding the primary function for interacting with ``cavcalc``
via the API.
"""

import numpy as np
import os as _os
from tempfile import NamedTemporaryFile as _NTF
from typing import Union as _Union, Sequence as _Sequence

from . import _SESSION, Q_
from .errors import CavCalcError
from ._handler import make_handler as _make_handler
from .parameters import ArgParameter, Parameter, ParameterType, valid_arguments, valid_targets
from .parameters.tools import get_name


def calculate(
    target: _Union[str, ParameterType] = None,
    meshes: _Union[bool, str, _Sequence[_Union[str, _Sequence[str]]]] = None,
    **kwargs,
):
    """Calculates a target parameter from an arbitrary number of physical arguments.

    If no `target` is specified then the default behaviour is to calculate all computable
    parameters from the given arguments.

    .. dropdown:: Valid targets and keyword arguments
        :icon: info
        :color: primary
        :animate: fade-in-slide-down

        You can also refer to the :ref:`param_ref` for a convenient overview of
        the information below.

        .. rubric:: Targets

        .. jupyter-execute::
            :hide-code:

            import cavcalc.parameters as ccp
            import pandas as pd

            table_styles = [
                {'selector': "th.col_heading", "props": 'text-align: center'},
                {"selector": "td", "props": "padding: 10px; text-align: center"},
                {
                    "selector": "th",
                    "props": "padding-left: 60px; padding-right: 60px; background-color: grey; text-align: center"
                },
                {"selector": "tr", "props": "border: solid; border-width: 0.5px 0"},
            ]

            pd.DataFrame(
                [
                    (p.name, p.description) for p in sorted(
                        (ccp.TargetParameter(name) for name in ccp.valid_targets),
                        key=lambda p: (p.category.value, p.ptype.value)
                    )
                ],
                columns=["Target", "Description"]
            ).style.hide().set_table_styles(table_styles)

        .. rubric:: Keyword arguments

        .. jupyter-execute::
            :hide-code:

            pd.DataFrame(
                [
                    (p.name, p.description) for p in sorted(
                        (ccp.ArgParameter(name) for name in ccp.valid_arguments),
                        key=lambda p: (p.category.value, p.ptype.value)
                    )
                ],
                columns=["Argument", "Description"]
            ).style.hide().set_table_styles(table_styles)

        Each kwarg value can be specified as:

        * A single value, or an array of values. The units of this value will correspond to
          those of the associated parameter, or parameter category, in the config file being used.
        * A ``cavcalc.ureg.Quantity`` instance. See the Pint documentation for details on
          instantiating ``Quantity`` objects.
        * A :class:`.Parameter` object: i.e. an entry from an existing :class:`.BaseOutput` instance.
        * Or any value which can be converted into a ``Quantity`` instance; e.g. a string
          (such as ``"10cm"``) or a tuple (such as ``(310, "deg")``).

    Parameters
    ----------
    target : str | :class:`.ParameterType`, optional
        The target parameter to compute, can be specified as a string (see drop-down box above) or
        a constant of the enum :class:`.ParameterType`. Defaults to ``None`` so that the function
        computes all the parameters it can from the given inputs.

    meshes : str | bool | Sequence[str | Sequence[str]], optional
        Parameter combinations from which to construct mesh-grids. This argument can given in
        a number of different ways, e.g:

        * `meshes=True`: constructs mesh-grids from the array-like arguments in the order in which
          they are given to this function.
        * `meshes="g1,g2"`: makes mesh-grids from arguments `g1` and `g2`, respectively.
        * `meshes=["g1,g2", "R1,R2"]`: makes mesh-grids from `g1` and `g2`, respectively, and also
          makes mesh-grids from `R1` and `R2`, respectively.
        * `meshes=("L, R1, R2")`: makes mesh-grids from `L1`, `R1` and `R2`, respectively.

    **kwargs : Keyword Arguments
        See the drop-down box above for details.

    Returns
    -------
    out : :class:`.SingleOutput` | :class:`.MultiOutput`
        An output object containing the results; with methods for accessing, displaying, and plotting them. If
        a ``target`` was given, and was not ``None``, then this object will be a :class:`.SingleOutput` instance,
        otherwise it will be a :class:`.MultiOutput` object.

    Examples
    --------
    The following imports are used for the below examples:

    .. jupyter-execute::
        :hide-output:

        import cavcalc as cc
        import numpy as np

        # This configures the matplotlib rcParams for the session
        # to use the style-sheet provided by cavcalc
        cc.configure("cavcalc")

    .. jupyter-execute::
        :hide-code:
        :hide-output:

        import matplotlib.pyplot as plt
        plt.rcParams["figure.figsize"] = [12, 7.416]
        plt.rcParams["font.size"] = 13

    Compute and show all determined properties from the cavity length and round-trip Gouy phase:

    .. jupyter-execute::

        print(cc.calculate(L=1, gouy=300))

    Calculate, and plot, the beam radii on the cavity mirrors over a range of radii of curvature:

    .. jupyter-execute::

        cc.calculate("w", L="4km", Rc=np.linspace(2.1e3, 2.5e3, 300)).plot();

    Make an image plot of the round-trip Gouy phase calculated on a grid over the stability
    factors of the cavity mirrors:

    .. jupyter-execute::

        g_arr = np.linspace(-2, 2, 399)
        # Setting meshes=True here will construct mesh-grids from all array
        # parameters in the order that they are specified
        cc.calculate("gouy", meshes=True, g1=g_arr, g2=g_arr).plot(cmap="Spectral_r");

    Get the radius of curvature of the cavity mirrors given the beam radii on them:

    .. jupyter-execute::

        out = cc.calculate(L="3cm", w1="100um", w2="120um")

        # .value is the pint.Quantity object
        print(f"Rc1 = {out.get('Rc1').value.to('mm'):~}")
        print(f"Rc2 = {out.get('Rc2').value.to('mm'):~}")

    Compute the fractional transmission intensity over a grid of the mirror reflectivities:

    .. jupyter-execute::

        cc.calculate("Atrn", R1=np.linspace(0, 1, 250), R2="R1", meshes=True).plot(cmap="hot");

    Plot the radius of the beam at the waist position, as a function of the beam radii
    on the mirrors; whilst using :func:`.configure` to temporarily override the default
    units (loaded from the config files) for beam-size type parameters:

    .. jupyter-execute::

        w_arr = np.linspace(80, 150, 499)
        with cc.configure(beamsizes="um"):
            fig = cc.calculate("w0", L="3cm", meshes=True, w1=w_arr, w2=w_arr).plot(show=False)
            fig.subplots_adjust(wspace=0.25)
    """
    if target is not None:
        if not isinstance(target, ParameterType):
            if target not in valid_targets:
                raise CavCalcError(f"Unrecognised / invalid target '{target}'")
        else:
            target = get_name(target)
            if target not in valid_targets:
                raise CavCalcError(f"Invalid target '{target}'")
    else:
        target = "all"

    tmp_files = []
    command = [target]
    for arg, value in kwargs.items():
        if arg not in valid_arguments:
            raise CavCalcError(f"Unrecognised / invalid argument '{arg}'")

        if isinstance(value, str) and value in valid_arguments:  # parameter reference
            command.extend([f"-{arg}", value])
            continue

        if isinstance(value, Parameter):
            value = value.value
        elif isinstance(value, tuple):
            if len(value) != 2:
                raise CavCalcError(f"Expected tuple of length 2 for argument '{arg}'")
            value = Q_(*value)
        else:
            value = Q_(value)

        arg_param = ArgParameter(arg, value)

        command.append(arg_param.cli_form)
        units_str = str(arg_param.value.units)
        if arg_param.is_array:
            # If the array is irregularly spaced, then we need to save it to a
            # tmp file for loading later on in the parser (as the data-range
            # syntax only allows for linearly spaced array args).
            if np.any(np.abs(np.diff(arg_param.value.m, n=2)) > 1e-13):
                with _NTF(suffix=".npy", delete=False) as f:
                    np.save(f, arg_param.value.m)
                    command.append(f"{f.name} {arg_param.value.u}")
                    tmp_files.append(f.name)
            else:
                start = np.min(arg_param.value.m)
                stop = np.max(arg_param.value.m)
                size = arg_param.value.m.size

                command.append(f"{start} {stop} {size} {units_str}")
        else:
            command.append(f"({arg_param.value.m}){units_str}")

    if meshes:
        command.append("--mesh")
        command.append(_deduce_meshes(meshes))

    out = _make_handler(_SESSION.parse_args(command)).run()

    for tmp_f in tmp_files:
        if _os.path.isfile(tmp_f):
            _os.remove(tmp_f)

    return out


def _deduce_meshes(meshes):
    if isinstance(meshes, (str, bool)):
        return str(meshes)

    err_msg = "Could not deduce mesh parameters from specified meshes argument."
    if not hasattr(meshes, "__getitem__"):
        raise CavCalcError(err_msg + " Expected this argument to be iterable.")

    if all((isinstance(m, str) and "," not in m) for m in meshes):
        return ",".join(meshes)

    mesh_str = ""
    for m in meshes:
        if isinstance(m, str):
            if "," not in m:
                raise CavCalcError(
                    err_msg + " Either all strings in this sequence must correspond to "
                    "single parameter names, or each string must represent a combination "
                    "of parameters separated by commas."
                )

            mesh_str += m
        else:
            if not hasattr(m, "__getitem__"):
                raise CavCalcError(
                    err_msg + " Expected each item in this sequence to be a string of "
                    "parameter name combinations separated by commas, or an iterable "
                    "of parameter names."
                )

            mesh_str += ",".join(m)

        mesh_str += ";"

    return mesh_str



from ..parameters import ParameterType

from .asymmetric import *
from .resonance import *
from .symmetric import *


RESONANCE_TARGETS_FUNC_MAP = {
    fsr: (
        (ParameterType.CAV_LENGTH,),
        ParameterType.FSR,
    ),
    R1_of_T1L1: (
        (ParameterType.TRANSMISSION_M1, ParameterType.LOSS_M1),
        ParameterType.REFLECTIVITY_M1,
    ),
    T1_of_R1L1: (
        (ParameterType.REFLECTIVITY_M1, ParameterType.LOSS_M1),
        ParameterType.TRANSMISSION_M1,
    ),
    R2_of_T2L2: (
        (ParameterType.TRANSMISSION_M2, ParameterType.LOSS_M2),
        ParameterType.REFLECTIVITY_M2,
    ),
    T2_of_R2L2: (
        (ParameterType.REFLECTIVITY_M2, ParameterType.LOSS_M2),
        ParameterType.TRANSMISSION_M2,
    ),
    finesse: (
        (ParameterType.REFLECTIVITY_M1, ParameterType.REFLECTIVITY_M2),
        ParameterType.FINESSE,
    ),
    fwhm: (
        (ParameterType.CAV_LENGTH, ParameterType.REFLECTIVITY_M1, ParameterType.REFLECTIVITY_M2),
        ParameterType.FWHM,
    ),
    pole: (
        (ParameterType.CAV_LENGTH, ParameterType.REFLECTIVITY_M1, ParameterType.REFLECTIVITY_M2),
        ParameterType.POLE,
    ),
    Aint_of_R1R2: (
        (ParameterType.REFLECTIVITY_M1, ParameterType.REFLECTIVITY_M2),
        ParameterType.RES_ENHANCE_INTERNAL,
    ),
    Aext_of_R1R2: (
        (ParameterType.REFLECTIVITY_M1, ParameterType.REFLECTIVITY_M2),
        ParameterType.RES_ENHANCE_EXTERNAL,
    ),
    Atrn_of_R1R2: (
        (ParameterType.REFLECTIVITY_M1, ParameterType.REFLECTIVITY_M2),
        ParameterType.TRANSMISSION_INTENSITY_FRAC,
    ),
}


SYMMETRIC_TARGETS_FUNC_MAP = {
    w_of_gsingle: (
        (ParameterType.CAV_LENGTH, ParameterType.WAVELENGTH, ParameterType.GFACTOR_SINGLE),
        ParameterType.BEAMSIZE,
    ),
    w0_of_gsingle: (
        (ParameterType.CAV_LENGTH, ParameterType.WAVELENGTH, ParameterType.GFACTOR_SINGLE),
        ParameterType.WAISTSIZE,
    ),
    z0_symmetric: (
        (ParameterType.CAV_LENGTH,),
        ParameterType.WAISTPOS,
    ),
    roc_of_gsingle: (
        (ParameterType.CAV_LENGTH, ParameterType.GFACTOR_SINGLE),
        ParameterType.ROC,
    ),
    gsingle_of_roc: ((ParameterType.CAV_LENGTH, ParameterType.ROC), ParameterType.GFACTOR_SINGLE),
    gsingle_of_w: (
        (ParameterType.CAV_LENGTH, ParameterType.WAVELENGTH, ParameterType.BEAMSIZE),
        ParameterType.GFACTOR_SINGLE,
    ),
    gsingle_of_rtgouy: ((ParameterType.GOUY,), ParameterType.GFACTOR_SINGLE),
    gsingle_of_divang: (
        (ParameterType.CAV_LENGTH, ParameterType.WAVELENGTH, ParameterType.DIVERGENCE),
        ParameterType.GFACTOR_SINGLE,
    ),
    gsingle_of_gcav: ((ParameterType.CAV_GFACTOR,), ParameterType.GFACTOR_SINGLE),
    gcav_of_gsingle: ((ParameterType.GFACTOR_SINGLE,), ParameterType.CAV_GFACTOR),
    rtgouy_of_gsingle: (
        (ParameterType.GFACTOR_SINGLE,),
        ParameterType.GOUY,
    ),
    divang_of_gsingle: (
        (ParameterType.CAV_LENGTH, ParameterType.WAVELENGTH, ParameterType.GFACTOR_SINGLE),
        ParameterType.DIVERGENCE,
    ),
    modesep_of_gsingle: (
        (ParameterType.CAV_LENGTH, ParameterType.GFACTOR_SINGLE),
        ParameterType.MODESEP,
    ),
}


ASYMMETRIC_TARGETS_FUNC_MAP = {
    w1_of_g1g2: (
        (
            ParameterType.CAV_LENGTH,
            ParameterType.WAVELENGTH,
            ParameterType.GFACTOR_M1,
            ParameterType.GFACTOR_M2,
        ),
        ParameterType.BEAMSIZE_M1,
    ),
    w2_of_g1g2: (
        (
            ParameterType.CAV_LENGTH,
            ParameterType.WAVELENGTH,
            ParameterType.GFACTOR_M1,
            ParameterType.GFACTOR_M2,
        ),
        ParameterType.BEAMSIZE_M2,
    ),
    w0_of_g1g2: (
        (
            ParameterType.CAV_LENGTH,
            ParameterType.WAVELENGTH,
            ParameterType.GFACTOR_M1,
            ParameterType.GFACTOR_M2,
        ),
        ParameterType.WAISTSIZE,
    ),
    z0_of_g1g2: (
        (ParameterType.CAV_LENGTH, ParameterType.GFACTOR_M1, ParameterType.GFACTOR_M2),
        ParameterType.WAISTPOS,
    ),
    rtgouy_of_g1g2: (
        (ParameterType.GFACTOR_M1, ParameterType.GFACTOR_M2),
        ParameterType.GOUY,
    ),
    divang_of_g1g2: (
        (
            ParameterType.CAV_LENGTH,
            ParameterType.WAVELENGTH,
            ParameterType.GFACTOR_M1,
            ParameterType.GFACTOR_M2,
        ),
        ParameterType.DIVERGENCE,
    ),
    g1_of_Rc1: (
        (ParameterType.CAV_LENGTH, ParameterType.ROC_M1),
        ParameterType.GFACTOR_M1,
    ),
    g2_of_Rc2: (
        (ParameterType.CAV_LENGTH, ParameterType.ROC_M2),
        ParameterType.GFACTOR_M2,
    ),
    g1_of_w1w2: (
        (
            ParameterType.CAV_LENGTH,
            ParameterType.WAVELENGTH,
            ParameterType.BEAMSIZE_M1,
            ParameterType.BEAMSIZE_M2,
        ),
        ParameterType.GFACTOR_M1,
    ),
    g2_of_w1w2: (
        (
            ParameterType.CAV_LENGTH,
            ParameterType.WAVELENGTH,
            ParameterType.BEAMSIZE_M1,
            ParameterType.BEAMSIZE_M2,
        ),
        ParameterType.GFACTOR_M2,
    ),
    g_of_g1g2: (
        (ParameterType.GFACTOR_M1, ParameterType.GFACTOR_M2),
        ParameterType.CAV_GFACTOR,
    ),
    Rc1_of_g1: (
        (ParameterType.CAV_LENGTH, ParameterType.GFACTOR_M1),
        ParameterType.ROC_M1,
    ),
    Rc2_of_g2: (
        (ParameterType.CAV_LENGTH, ParameterType.GFACTOR_M2),
        ParameterType.ROC_M2,
    ),
    modesep_of_g1g2: (
        (ParameterType.CAV_LENGTH, ParameterType.GFACTOR_M1, ParameterType.GFACTOR_M2),
        ParameterType.MODESEP,
    ),
}



from functools import wraps
import numpy as np
import warnings

from .._exiters import quit_print


def check_physical():
    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                    result = function(*args, **kwargs)
            except ZeroDivisionError:
                target = function.__name__.split("_")[0]
                quit_print(f"Division by zero occurred when computing target: '{target}'")

            return result

        return wrapper

    return decorator


def modesep_adjust(rtgouy, L):
    """Compute the mode-separation frequency taking into account
    the value of the round-trip Gouy phase.
    """
    from .resonance import _fsr_base

    gouy = np.atleast_1d(rtgouy)
    df = (0.5 * gouy / np.pi) * _fsr_base(L)
    df = np.where(gouy > np.pi, _fsr_base(L) - df, df)

    if not isinstance(rtgouy, np.ndarray) and not isinstance(L, np.ndarray):
        return df[0]

    return df



from pint.errors import DimensionalityError, UndefinedUnitError

from . import _CONFIG, Q_
from ._arguments._container import _Arguments
from .errors import CavCalcError
from ._exiters import quit_print, bug
from .functions._maps import (
    RESONANCE_TARGETS_FUNC_MAP as _RES_TARGETS_MAP,
    SYMMETRIC_TARGETS_FUNC_MAP as _SYMM_TARGETS_MAP,
    ASYMMETRIC_TARGETS_FUNC_MAP as _ASYMM_TARGETS_MAP,
)
from .output import SingleOutput, MultiOutput
from .parameters import ArgParameter, TargetParameter, ParameterType
from .parameters.tools import get_name


def make_handler(args: _Arguments):
    if args.is_single_target:
        handler = _SingleTargetHandler(args)
    else:
        handler = _AllTargetHandler(args)

    return handler


class _Handler:
    def __init__(self, args: _Arguments):
        self.args = args

        is_symmetric = not any(
            (self.args.has(f"{name}1") or self.args.has(f"{name}2")) for name in ("w", "Rc", "g")
        )

        # Make a copy to ensure we do not modify resonance's dependencies map itself!
        self.func_dep_map = _RES_TARGETS_MAP.copy()
        if is_symmetric:
            self.func_dep_map |= _SYMM_TARGETS_MAP
        else:
            self.func_dep_map |= _ASYMM_TARGETS_MAP

    def _get_computable_funcs(self):
        all_poss_computable_params = set(rp for _, rp in self.func_dep_map.values())

        def _update_computables(phys_arg_params: set[ParameterType]):
            computables = {}
            for target in all_poss_computable_params:
                if target in phys_arg_params or target in computables:
                    continue

                for func, (reqd_params, ret_param) in self.func_dep_map.items():
                    if ret_param == target:
                        if not set(reqd_params).difference(phys_arg_params):
                            if computables.get(ret_param):
                                quit_print(
                                    "Incompatible arguments detected! The parameter "
                                    f"'{get_name(ret_param)}' can be computed multiple different ways "
                                    "based on the arguments given."
                                )

                            computables[ret_param] = func

            return computables

        computable_funcs = {}
        available_params = set(self.args.physical_args.keys())
        # Keep updating the physical argument parameter set using
        # the functions we just determined could be computed. This
        # let's us chain together many targets from a relatively
        # sparse initial set of given arguments.
        while new_funcs := _update_computables(available_params):
            computable_funcs |= new_funcs
            available_params |= new_funcs.keys()

        return computable_funcs


class _SingleTargetHandler(_Handler):
    def __init__(self, args: _Arguments):
        super().__init__(args)
        phys_arg_params = set(self.args.physical_args.keys())
        self.target_parameter = TargetParameter(self.args.compute)

        # Get the functions which have target parameter as return
        target_functions = {
            func: (reqd_params, ret_param)
            for func, (reqd_params, ret_param) in self.func_dep_map.items()
            if ret_param == self.target_parameter.ptype
        }

        def _find_target_func():
            return next(
                (
                    func
                    for func, (reqd_params, _) in target_functions.items()
                    if not set(reqd_params).difference(phys_arg_params)
                ),
                None,
            )

        self.target_function = _find_target_func()
        if self.target_function is None:  # couldn't compute it directly, so let's do chaining
            # First step of chaining logic is to get the functions that we *can*
            # compute from both our initial arg set and the extra chained args
            # themselves as we add them
            computable_funcs = self._get_computable_funcs()
            phys_arg_params |= computable_funcs.keys()

            self.target_function = _find_target_func()
            if self.target_function:
                reqd_params, _ = self.func_dep_map[self.target_function]
                # Which chained targets do we need to compute?
                diff = set(reqd_params).difference(self.args.physical_args.keys())
                self.chained = {}
                for pd in diff:  # calculate each of these in-turn...
                    chained_func = computable_funcs[pd]
                    chained_reqd, _ = self.func_dep_map[chained_func]
                    chained_func_args = tuple(
                        self.args.physical_args.get(p) or self.chained.get(p) for p in chained_reqd
                    )
                    # ... storing them as chained ArgParameter instances, for
                    # easy use when calculating self.target_function later on
                    self.chained[pd] = ArgParameter(pd, compute(chained_func, *chained_func_args))

                    for arg in chained_func_args:
                        arg._used = True
            else:
                msg = f"Incorrect usage. To compute {self.args.compute} I require one of: "
                all_maps = _RES_TARGETS_MAP | _SYMM_TARGETS_MAP | _ASYMM_TARGETS_MAP
                for reqd_params, ret_param in all_maps.values():
                    if ret_param == self.target_parameter.ptype:
                        msg += "\n\t"
                        msg += " AND ".join(
                            ArgParameter(param).cli_form
                            if param != ParameterType.WAVELENGTH
                            else f"[-wl = {self.args.wl.value:~}]"
                            for param in reqd_params
                        )

                msg += (
                    "\n\nOR a combination of parameters which allows one of "
                    "the above sets to be computed."
                )
                quit_print(msg)

    def run(self):
        reqd_params, _ = self.func_dep_map[self.target_function]
        func_args = tuple(
            self.args.physical_args.get(p) or self.chained.get(p) for p in reqd_params
        )

        result = compute(self.target_function, *func_args)
        units = self.args["units"] or _CONFIG["units"].get(self.target_parameter.name)
        if not self.target_parameter.is_unitless:
            try:
                result = result.to(units)
            except (DimensionalityError, UndefinedUnitError, ValueError) as ex:
                quit_print(str(ex))
        else:
            if units:
                quit_print(
                    f"Units '{units}' given for dimensionless target '{self.target_parameter.name}'"
                )

        for arg in func_args:
            arg._used = True

        self.target_parameter = TargetParameter(self.target_parameter.name, result)
        return SingleOutput(self.target_parameter, tuple(self.args.physical_args.values()))


class _AllTargetHandler(_Handler):
    def __init__(self, args):
        super().__init__(args)

        # Get all the targets that we can compute from both the initial
        # argument set and the extra chained targets
        self.target_functions = tuple(self._get_computable_funcs().values())

    def run(self):
        computed_targets: dict[TargetParameter, tuple[ParameterType]] = {}
        for func in self.target_functions:
            reqd_params, ret_param = self.func_dep_map[func]
            func_args = tuple(
                (
                    self.args.physical_args.get(p)
                    or next((arg for arg in computed_targets.keys() if arg.ptype == p))
                )
                for p in reqd_params
            )

            target_param = TargetParameter(ret_param)

            result = compute(func, *func_args)
            units = _CONFIG["units"].get(target_param.name)
            if not target_param.is_unitless:
                try:
                    result = result.to(units)
                except (DimensionalityError, UndefinedUnitError, ValueError) as ex:
                    quit_print(str(ex))
            else:
                if units:
                    quit_print(
                        f"Units '{units}' given for dimensionless target '{target_param.name}'"
                    )

            for arg in filter(lambda p: isinstance(p, ArgParameter), func_args):
                arg._used = True

            target_param = TargetParameter(ret_param, result)
            computed_targets[target_param] = reqd_params

        to_update = {}
        for tgt_p, reqd_params in computed_targets.items():
            for rp in reqd_params:
                # A required parameter was not specified, so it must be a chained target...
                if rp not in self.args.physical_args:
                    # ... find this target and get *its* required params
                    chained_target = next((tp for tp in computed_targets if tp.ptype == rp), None)
                    if not chained_target:
                        bug(f"Could not find chained target parameter corresponding to {rp}")

                    chained_tgt_reqd_params = computed_targets[chained_target]
                    # Now get the chained target required params which *were* specified...
                    extra_params = tuple(
                        filter(lambda p: p in self.args.physical_args, chained_tgt_reqd_params)
                    )
                    # ... and add them as chained extra required params
                    if extra_params:
                        if tgt_p in to_update:
                            to_update[tgt_p] += extra_params
                        else:
                            to_update[tgt_p] = extra_params

        # These chained target parameter types are then added to the relevant
        # target mappings as extra "required" params, so that we can correctly
        # construct SingleOutput instances from within the MultiOutput
        for tgt_p, extras in to_update.items():
            computed_targets[tgt_p] = tuple(set(computed_targets[tgt_p] + extras))

        return MultiOutput(computed_targets, tuple(self.args.physical_args.values()))


def compute(func, *args) -> Q_:
    try:
        return func(*(arg_p.value for arg_p in args))
    except Exception as ex:
        if isinstance(ex, CavCalcError):
            raise

        if isinstance(ex, ValueError) and "operands could not be broadcast" in str(ex):
            shapes = (str(getattr(arg_p.value.m, "shape", 1)) for arg_p in args)
            quit_print(
                f"Encountered inconsistent parameter shapes when calling {func.__name__}. "
                f"Parameters: {{{', '.join(arg_p.name for arg_p in args)}}}, with shapes: "
                f"{{{', '.join(shapes)}}}, could not be broadcast together."
            )
        else:
            bug(
                "The following error occurred:\n\n"
                f"\t{ex}\n\n"
                f"when calling {func.__name__} with arguments:\n\n"
                + "\n\t".join(str(arg_p) for arg_p in args)
            )



"""
Physical geometric-like parameter functions associated with symmetric optical cavities. Most properties
here are expressed only in terms of the g-factor of the cavity mirrors, as inter-dependent relations
between these properties are computed via target chaining in the core cavcalc code.

.. note::

    Whilst these underlying functions are publicly accessible, it is recommended that you
    instead use the single-function interface :func:`.calculate` for computing any
    of the cavity properties that you need; as that will result in a nice output
    object for accessing results.
"""

import numpy as np

from .. import ureg
from ._utils import check_physical, modesep_adjust


@check_physical()
@ureg.wraps(ureg.meter, (ureg.meter, ureg.meter, ""))
def w_of_gsingle(L, wl, gsingle):
    r"""Radius of a beam, of wavelength :math:`\lambda`, on both mirrors of a symmetric cavity
    of length :math:`L` with mirror g-factors of :math:`g_s`.

    .. math::
        w\left(L, \lambda, g_s\right) = \sqrt{
            \frac{L \lambda}{\pi} \sqrt{\frac{1}{1 - g_s^2}}
        }
    """
    return np.sqrt(L * wl / np.pi * np.sqrt(1 / (1 - gsingle * gsingle)))


@check_physical()
@ureg.wraps(ureg.meter, (ureg.meter, ureg.meter, ""))
def w0_of_gsingle(L, wl, gsingle):
    r"""Radius of a beam, of wavelength :math:`\lambda`, at the waist position of a symmetric
    cavity of length :math:`L` with mirror g-factors of :math:`g_s`.

    .. math::
        w_0\left(L, \lambda, g_s\right) = \sqrt{
            \frac{L \lambda}{2\pi} \sqrt{\frac{1 + g_s}{1 - g_s}}
        }
    """
    return np.sqrt(0.5 * L * wl / np.pi * np.sqrt((1 + gsingle) / (1 - gsingle)))


@check_physical()
@ureg.wraps(ureg.meter, ureg.meter)
def z0_symmetric(L):
    r"""Waist position of a beam in a symmetric cavity of length :math:`L`.

    .. math::
        z_0(L) = \frac{L}{2}
    """
    return 0.5 * L


@check_physical()
@ureg.wraps(ureg.meter, (ureg.meter, ""))
def roc_of_gsingle(L, gsingle):
    r"""Radius of curvature of both cavity mirrors in a symmetric cavity of length :math:`L`,
    with mirror g-factors of :math:`g_s`.

    .. math::
        R_C(L, g_s) = \frac{L}{1 - g_s}
    """
    return L / (1 - gsingle)


@check_physical()
@ureg.wraps("", (ureg.meter, ureg.meter))
def gsingle_of_roc(L, Rc):
    r"""Mirror stability factor given radius of curvature :math:`R_C`, in a symmetric cavity
    of length :math:`L`.

    .. math::
        g_s(L, R_C) = 1 - \frac{L}{R_C}
    """
    return 1 - L / Rc


@check_physical()
@ureg.wraps("", (ureg.meter, ureg.meter, ureg.meter))
def gsingle_of_w(L, wl, w):
    r"""Mirror stability factor given beam radius (on both mirrors) of :math:`w`, in a symmetric
    cavity of length :math:`L`; where the beam has wavelength :math:`\lambda`.

    .. math::
        g_s(L, \lambda, w) = \pm \sqrt{
            1 - \left( \frac{L\lambda}{\pi w^2} \right)^2
        }
    """
    mag = np.sqrt(1 - (L * wl / (np.pi * w * w)) ** 2)
    return np.array([-mag, mag])


@check_physical()
@ureg.wraps("", ureg.radians)
def gsingle_of_rtgouy(rtgouy):
    r"""Mirror stability factor given round-trip gouy phase of :math:`\psi`, of a symmetric cavity.

    .. math::
        g_s(\psi) = \cos{\left(\frac{\psi}{2}\right)}
    """
    return np.cos(0.5 * rtgouy)


@check_physical()
@ureg.wraps("", (ureg.meter, ureg.meter, ureg.radians))
def gsingle_of_divang(L, wl, theta):
    r"""Mirror stability factor for a symmetric cavity of length :math:`L` with an eigenmode
    with divergence angle :math:`\theta`; where the beam has wavelength :math:`\lambda`.

    .. math::
        g_s(L, \lambda, \theta) = \frac{
            1 - \left(\frac{L\pi}{2\lambda}\right)^2 \tan^4{\theta}
        }{
            1 + \left(\frac{L\pi}{2\lambda}\right)^2 \tan^4{\theta}
        }
    """
    factor_L_wl = (L * np.pi / (2 * wl)) ** 2 * np.power(np.tan(theta), 4)
    return (1 - factor_L_wl) / (1 + factor_L_wl)


@check_physical()
@ureg.wraps("", (""))
def gsingle_of_gcav(gcav):
    r"""Mirror stability factor for a symmetric cavity with overall g-factor :math:`g`.

    .. math::
        g_s(g) = \pm \sqrt{g}
    """
    mag = np.sqrt(gcav)
    return np.array([-mag, mag])


@check_physical()
@ureg.wraps("", "")
def gcav_of_gsingle(gsingle):
    r"""Symmetric cavity stability factor given mirror g-factors of :math:`g_s`.

    .. math::
        g(g_s) = g_s^2
    """
    return gsingle * gsingle


def _rtgouy_of_gsingle_base(gsingle):
    return 2 * np.arccos(gsingle)


@check_physical()
@ureg.wraps(ureg.radians, "")
def rtgouy_of_gsingle(gsingle):
    r"""Round-trip gouy phase of a symmetric cavity with mirror g-factors of :math:`g_s`.

    .. math::
        \psi(g_s) = 2 \arccos{\left(g_s\right)}
    """
    return _rtgouy_of_gsingle_base(gsingle)


@check_physical()
@ureg.wraps(ureg.radians, (ureg.meter, ureg.meter, ""))
def divang_of_gsingle(L, wl, gsingle):
    r"""Divergence angle of the beam in a symmetric cavity of length :math:`L`, with
    mirror g-factors of :math:`g_s`; where the beam has wavelength :math:`\lambda`.

    .. math::
        \theta(L, \lambda, g_s) = \arctan{\left(\sqrt{
            \frac{2\lambda}{L\pi}
            \sqrt{\frac{1 - g_s}{1 + g_s}}
        }\right)}
    """
    return np.arctan(np.sqrt(2 * wl / (L * np.pi) * np.sqrt((1 - gsingle) / (1 + gsingle))))


@check_physical()
@ureg.wraps(ureg.hertz, (ureg.meter, ""))
def modesep_of_gsingle(L, gsingle):
    r"""Mode separation frequency for a symmetric cavity of length :math:`L`, with
    mirror g-factors of :math:`g_s`.

    In the equation below, :math:`\nu(L)` is the FSR of the cavity, and :math:`\psi(g_s)`
    is the round-trip gouy phase.

    .. math::
        \delta f(L, g_s) = \begin{cases}
            \frac{\psi(g_s)}{2\pi}\,\nu(L) \quad \text{if }\, \psi < \pi,\\
            \nu(L)\left(1 - \frac{\psi(g_s)}{2\pi}\right) \quad \text{otherwise}
        \end{cases}
    """
    rtgouy = _rtgouy_of_gsingle_base(gsingle)
    return modesep_adjust(rtgouy, L)



from . import ParameterType, ParameterCategory


NAME_PTYPE_MAP = {
    "w": ParameterType.BEAMSIZE,
    "w1": ParameterType.BEAMSIZE_M1,
    "w2": ParameterType.BEAMSIZE_M2,
    "g": ParameterType.CAV_GFACTOR,
    "L": ParameterType.CAV_LENGTH,
    "div": ParameterType.DIVERGENCE,
    "finesse": ParameterType.FINESSE,
    "FSR": ParameterType.FSR,
    "FWHM": ParameterType.FWHM,
    "g1": ParameterType.GFACTOR_M1,
    "g2": ParameterType.GFACTOR_M2,
    "gouy": ParameterType.GOUY,
    "gs": ParameterType.GFACTOR_SINGLE,
    "L1": ParameterType.LOSS_M1,
    "L2": ParameterType.LOSS_M2,
    "modesep": ParameterType.MODESEP,
    "pole": ParameterType.POLE,
    "R1": ParameterType.REFLECTIVITY_M1,
    "R2": ParameterType.REFLECTIVITY_M2,
    "Aint": ParameterType.RES_ENHANCE_INTERNAL,
    "Aext": ParameterType.RES_ENHANCE_EXTERNAL,
    "Rc": ParameterType.ROC,
    "Rc1": ParameterType.ROC_M1,
    "Rc2": ParameterType.ROC_M2,
    "Atrn": ParameterType.TRANSMISSION_INTENSITY_FRAC,
    "T1": ParameterType.TRANSMISSION_M1,
    "T2": ParameterType.TRANSMISSION_M2,
    "z0": ParameterType.WAISTPOS,
    "w0": ParameterType.WAISTSIZE,
    "wl": ParameterType.WAVELENGTH,
}

PTYPE_NAME_MAP = {ptype: pname for pname, ptype in NAME_PTYPE_MAP.items()}

PTYPE_CATEGORY_MAP = {
    ParameterType.DIVERGENCE: ParameterCategory.Angle,
    ParameterType.BEAMSIZE: ParameterCategory.BeamRadius,
    ParameterType.BEAMSIZE_M1: ParameterCategory.BeamRadius,
    ParameterType.BEAMSIZE_M2: ParameterCategory.BeamRadius,
    ParameterType.WAISTSIZE: ParameterCategory.BeamRadius,
    ParameterType.ROC: ParameterCategory.Curvature,
    ParameterType.ROC_M1: ParameterCategory.Curvature,
    ParameterType.ROC_M2: ParameterCategory.Curvature,
    ParameterType.CAV_LENGTH: ParameterCategory.Distance,
    ParameterType.WAISTPOS: ParameterCategory.Distance,
    ParameterType.FSR: ParameterCategory.Frequency,
    ParameterType.FWHM: ParameterCategory.Frequency,
    ParameterType.POLE: ParameterCategory.Frequency,
    ParameterType.MODESEP: ParameterCategory.Frequency,
    ParameterType.GOUY: ParameterCategory.Phase,
    ParameterType.LOSS_M1: ParameterCategory.Power,
    ParameterType.LOSS_M2: ParameterCategory.Power,
    ParameterType.REFLECTIVITY_M1: ParameterCategory.Power,
    ParameterType.REFLECTIVITY_M2: ParameterCategory.Power,
    ParameterType.RES_ENHANCE_INTERNAL: ParameterCategory.Power,
    ParameterType.RES_ENHANCE_EXTERNAL: ParameterCategory.Power,
    ParameterType.TRANSMISSION_INTENSITY_FRAC: ParameterCategory.Power,
    ParameterType.TRANSMISSION_M1: ParameterCategory.Power,
    ParameterType.TRANSMISSION_M2: ParameterCategory.Power,
    ParameterType.FINESSE: ParameterCategory.Power,
    ParameterType.CAV_GFACTOR: ParameterCategory.Stability,
    ParameterType.GFACTOR_SINGLE: ParameterCategory.Stability,
    ParameterType.GFACTOR_M1: ParameterCategory.Stability,
    ParameterType.GFACTOR_M2: ParameterCategory.Stability,
    ParameterType.WAVELENGTH: ParameterCategory.Wave,
}

PTYPE_DESCR_MAP = {
    ParameterType.BEAMSIZE: "Radius of beam at mirrors",
    ParameterType.BEAMSIZE_M1: "Radius of beam at first mirror",
    ParameterType.BEAMSIZE_M2: "Radius of beam at second mirror",
    ParameterType.CAV_GFACTOR: "Stability g-factor of cavity",
    ParameterType.CAV_LENGTH: "Cavity length",
    ParameterType.DIVERGENCE: "Divergence angle",
    ParameterType.FINESSE: "Finesse",
    ParameterType.FSR: "FSR",
    ParameterType.FWHM: "FWHM",
    ParameterType.GFACTOR_M1: "Stability g-factor of first mirror",
    ParameterType.GFACTOR_M2: "Stability g-factor of second mirror",
    ParameterType.GFACTOR_SINGLE: "Stability g-factor of both mirrors",
    ParameterType.GOUY: "Round-trip Gouy phase",
    ParameterType.LOSS_M1: "Loss of first mirror",
    ParameterType.LOSS_M2: "Loss of second mirror",
    ParameterType.MODESEP: "Mode separation frequency",
    ParameterType.POLE: "Pole frequency",
    ParameterType.REFLECTIVITY_M1: "Reflectivity of first mirror",
    ParameterType.REFLECTIVITY_M2: "Reflectivity of second mirror",
    ParameterType.RES_ENHANCE_INTERNAL: "Internal resonance enhancement factor",
    ParameterType.RES_ENHANCE_EXTERNAL: "External resonance enhancement factor",
    ParameterType.ROC: "Radius of curvature of both mirrors",
    ParameterType.ROC_M1: "Radius of curvature of first mirror",
    ParameterType.ROC_M2: "Radius of curvature of second mirror",
    ParameterType.TRANSMISSION_INTENSITY_FRAC: "Fractional transmission intensity",
    ParameterType.TRANSMISSION_M1: "Transmission of first mirror",
    ParameterType.TRANSMISSION_M2: "Transmission of second mirror",
    ParameterType.WAISTPOS: "Position of beam waist (from first mirror)",
    ParameterType.WAISTSIZE: "Radius of beam at waist",
    ParameterType.WAVELENGTH: "Wavelength of beam",
}

PTYPE_SYMBOL_MAP = {
    ParameterType.BEAMSIZE: "$w$",
    ParameterType.BEAMSIZE_M1: "$w_1$",
    ParameterType.BEAMSIZE_M2: "$w_2$",
    ParameterType.CAV_GFACTOR: "$g$",
    ParameterType.CAV_LENGTH: "$L$",
    ParameterType.DIVERGENCE: r"$\theta$",
    ParameterType.FINESSE: r"$\mathcal{F}$",
    ParameterType.FSR: r"$\Delta \nu_{\mathrm{FSR}}$",
    ParameterType.FWHM: "FWHM",
    ParameterType.GFACTOR_M1: "$g_1$",
    ParameterType.GFACTOR_M2: "$g_2$",
    ParameterType.GFACTOR_SINGLE: "$g_s$",
    ParameterType.GOUY: r"$\psi$",
    ParameterType.LOSS_M1: "$L_1$",
    ParameterType.LOSS_M2: "$L_2$",
    ParameterType.MODESEP: r"$\delta f$",
    ParameterType.POLE: r"$\nu_p$",
    ParameterType.REFLECTIVITY_M1: "$R_1$",
    ParameterType.REFLECTIVITY_M2: "$R_2$",
    ParameterType.RES_ENHANCE_INTERNAL: r"$A_{\mathrm{circ}}$",
    ParameterType.RES_ENHANCE_EXTERNAL: r"$A'_{\mathrm{circ}}$",
    ParameterType.ROC: "$R_C$",
    ParameterType.ROC_M1: "$R_{C,1}$",
    ParameterType.ROC_M2: "$R_{C,2}$",
    ParameterType.TRANSMISSION_INTENSITY_FRAC: r"$A'_{\mathrm{trn}}$",
    ParameterType.TRANSMISSION_M1: "$T_1$",
    ParameterType.TRANSMISSION_M2: "$T_2$",
    ParameterType.WAISTPOS: "$z_0$",
    ParameterType.WAISTSIZE: "$w_0$",
    ParameterType.WAVELENGTH: r"$\lambda$",
}



"""
Optical resonance type properties of a Fabry-Perot cavity.

.. note::

    Whilst these underlying functions are publicly accessible, it is recommended that you
    instead use the single-function interface :func:`.calculate` for computing any
    of the cavity properties that you need; as that will result in a nice output
    object for accessing results.
"""

import numpy as np

from .. import ureg
from ._utils import check_physical


_C_LIGHT = 299792458.0


def _fsr_base(L):
    return 0.5 * _C_LIGHT / L


@check_physical()
@ureg.wraps(ureg.hertz, ureg.meter)
def fsr(L):
    r"""Free Spectral Range (FSR) of a cavity of length :math:`L`.

    In the equation below, :math:`c` is the speed of light in a vacuum.

    .. math::
        \nu(L) = \frac{c}{2L}
    """
    return _fsr_base(L)


@check_physical()
@ureg.wraps("", ("", ""))
def R1_of_T1L1(T1, L1):
    r"""Reflectivity of the first cavity mirror, where this mirror has
    a transmission of :math:`T_1` and a loss value of :math:`L_1`.

    .. math::
        R_1 = 1 - T_1 - L_1
    """
    return 1 - T1 - L1


@check_physical()
@ureg.wraps("", ("", ""))
def T1_of_R1L1(R1, L1):
    r"""Transmission of the first cavity mirror, where this mirror has
    a reflectivity of :math:`R_1` and a loss value of :math:`L_1`.

    .. math::
        T_1 = 1 - R_1 - L_1
    """
    return 1 - R1 - L1


@check_physical()
@ureg.wraps("", ("", ""))
def R2_of_T2L2(T2, L2):
    r"""Reflectivity of the second cavity mirror, where this mirror has
    a transmission of :math:`T_2` and a loss value of :math:`L_2`.

    .. math::
        R_2 = 1 - T_2 - L_2
    """
    return 1 - T2 - L2


@check_physical()
@ureg.wraps("", ("", ""))
def T2_of_R2L2(R2, L2):
    r"""Transmission of the second cavity mirror, where this mirror has
    a reflectivity of :math:`R_2` and a loss value of :math:`L_2`.

    .. math::
        T_2 = 1 - R_2 - L_2
    """
    return 1 - R2 - L2


def _finesse_base(R1, R2):
    return 0.5 * np.pi / np.arcsin(0.5 * (1 - np.sqrt(R1 * R2)) / (R1 * R2) ** 0.25)


@check_physical()
@ureg.wraps("", ("", ""))
def finesse(R1, R2):
    r"""Finesse of a cavity where the first mirror has (power) reflectivity :math:`R_1`
    and the second mirror has reflectivity :math:`R_2`.

    .. math::
        \mathcal{F}(R_1, R_2) = \frac{\pi}{
            2 \arcsin{
                \left( \frac{1 - \sqrt{R_1 R_2}}{2 (R_1 R_2)^{1/4}} \right)
            }
        }
    """
    return _finesse_base(R1, R2)


@check_physical()
@ureg.wraps(ureg.hertz, (ureg.meter, "", ""))
def fwhm(L, R1, R2):
    r"""Full-Width at Half-Maximum (FWHM) frequency of a cavity of length :math:`L`, where
    the first mirror has (power) reflectivity :math:`R_1` and the second mirror has
    reflectivity :math:`R_2`.

    In the equation below, :math:`\nu(L)` is the FSR, and :math:`\mathcal{F}(R_1,R_2)` is
    the finesse.

    .. math::
        \mathrm{FWHM}(L, R_1, R_2) = \frac{\nu(L)}{\mathcal{F}(R_1, R_2)}
    """
    return _fsr_base(L) / _finesse_base(R1, R2)


@check_physical()
@ureg.wraps(ureg.hertz, (ureg.meter, "", ""))
def pole(L, R1, R2):
    r"""The pole frequency of a cavity of length :math:`L`, where the first mirror has (power)
    reflectivity :math:`R_1` and the second mirror has reflectivity :math:`R_2`.

    In the equation below, :math:`\nu(L)` is the FSR, and :math:`\mathcal{F}(R_1,R_2)` is
    the finesse.

    .. math::
        \nu_p(L, R_1, R_2) = \frac{\nu(L)}{2\mathcal{F}(R_1, R_2)}
    """
    return 0.5 * _fsr_base(L) / _finesse_base(R1, R2)


@check_physical()
@ureg.wraps("", ("", ""))
def Aint_of_R1R2(R1, R2):
    r"""The internal resonance enhancement factor of a cavity where the first mirror has
    (power) reflectivity :math:`R_1` and the second mirror has reflectivity :math:`R_2`.

    .. math::
        A_{\mathrm{circ}} = \frac{1}{(1 - \sqrt{R_1 R_2})^2}
    """
    return 1 / (1 - np.sqrt(R1 * R2)) ** 2


@check_physical()
@ureg.wraps("", ("", ""))
def Aext_of_R1R2(R1, R2):
    r"""The external resonance enhancement factor of a cavity where the first mirror has
    (power) reflectivity :math:`R_1` and the second mirror has reflectivity :math:`R_2`.

    .. math::
        A'_{\mathrm{circ}} = \frac{1 - R_1}{(1 - \sqrt{R_1 R_2})^2}
    """
    return (1 - R1) / (1 - np.sqrt(R1 * R2)) ** 2


@check_physical()
@ureg.wraps("", ("", ""))
def Atrn_of_R1R2(R1, R2):
    r"""The fractional transmission intensity of a cavity where the first mirror has
    (power) reflectivity :math:`R_1` and the second mirror has reflectivity :math:`R_2`.

    .. math::
        A'_{\mathrm{trn}} = \frac{(1 - R_1)(1 - R_2)}{(1 - \sqrt{R_1 R_2})^2}
    """
    return (1 - R1) * (1 - R2) / (1 - np.sqrt(R1 * R2)) ** 2



"""
Generic classes for holding information on any cavcalc parameter type.
"""

import abc
import numpy as np
from typing import Union as _Union

from .. import Q_
from .._exiters import (
    bug as _bug,
    quit_print as _quit_print,
)
from . import ParameterType, ParameterCategory
from ._maps import (
    NAME_PTYPE_MAP as _NAME_PTYPE_MAP,
    PTYPE_CATEGORY_MAP as _PTYPE_CATEGORY_MAP,
    PTYPE_DESCR_MAP as _PTYPE_DESCR_MAP,
    PTYPE_NAME_MAP as _PTYPE_NAME_MAP,
    PTYPE_SYMBOL_MAP as _PTYPE_SYMBOL_MAP,
)
from .tools import get_default_units


class Parameter(abc.ABC):
    """Abstract base class for cavity parameter instances."""

    def __init__(self, name_or_ptype: _Union[str, ParameterType], value: Q_ = None):
        if isinstance(name_or_ptype, str):
            self.__name = name_or_ptype
            self.__ptype = _NAME_PTYPE_MAP.get(self.name)
            if self.ptype is None:
                _bug(f"No ParameterType associated with name: {self.name}")
        elif isinstance(name_or_ptype, ParameterType):
            self.__ptype = name_or_ptype
            self.__name = _PTYPE_NAME_MAP.get(self.ptype)
            if self.name is None:
                _bug(f"No name associated with ParameterType: {self.ptype}")
        else:
            _bug(f"An object, {name_or_ptype}, of unrecognised type was passed to Parameter.")

        self.__category = _PTYPE_CATEGORY_MAP.get(self.ptype)
        if self.category is None:
            _bug(f"No ParameterCategory associated with name: {self.name}")

        if value is not None:
            if isinstance(self, ArgParameter):
                default_units = get_default_units(self.name) if not self.is_unitless else ""

                # For any quantity specified without units, choose the default units
                if value.unitless and not self.is_unitless:
                    value = Q_(value.m, default_units)

                if not value.check(default_units):
                    _quit_print(f"Invalid units ({value.units}) given for parameter {self.name}")

        self._value = value

    def __str__(self):
        return self.description + (
            f" = {self.value:~}" if self.value is not None else " [uninitialized]"
        )

    @property
    def category(self) -> ParameterCategory:
        """The :class:`.ParameterCategory` that this parameter belongs to."""
        return self.__category

    @property
    def description(self):
        """A brief description of the parameter type."""
        return _PTYPE_DESCR_MAP[self.ptype]

    @property
    def has_angular_units(self):
        """Whether the units of this parameter are of angular type."""
        cat = self.category
        PC = ParameterCategory
        return cat == PC.Angle or cat == PC.Phase

    @property
    def has_frequency_units(self):
        """Whether the units of this parameter are of frequency type."""
        return self.category == ParameterCategory.Frequency

    @property
    def has_length_units(self):
        """Whether the units of this parameter are of length type.."""
        PC = ParameterCategory
        return self.category in (PC.BeamRadius, PC.Curvature, PC.Distance, PC.Wave)

    @property
    def is_array(self):
        """Whether the magnitude of the current value, if any, is array-like."""
        return self.value is not None and isinstance(self.value.m, np.ndarray)

    @property
    def is_scalar(self):
        """Whether the magnitude of the current value, if any, is not array-like."""
        return self.value is not None and not self.is_array

    @property
    def is_unitless(self):
        """Whether this parameter is a dimensionless type."""
        cat = self.category
        PC = ParameterCategory
        return cat == PC.Power or cat == PC.Stability

    @property
    def name(self) -> str:
        """The name of the parameter."""
        return self.__name

    @property
    def ptype(self) -> ParameterType:
        """The type of the parameter."""
        return self.__ptype

    @property
    def symbol_str(self):
        """A LaTeX style string representation of the parameter name."""
        return _PTYPE_SYMBOL_MAP[self.ptype]

    @property
    def value(self) -> Q_:
        """The current value of the parameter, as a pint Quantity."""
        return self._value


class ArgParameter(Parameter):
    """Cavity parameters which represent arguments given to cavcalc.

    .. note::

        You should never need to instantiate objects of this type manually. These
        are constructed during the argument parsing, and calculation process.
    """

    def __init__(
        self,
        name_or_ptype: _Union[str, ParameterType],
        value: Q_ = None,
        index: int = None,
        axis: int = None,
    ):
        super().__init__(name_or_ptype, value)
        self.__index = index
        self.__axis = axis
        self._used = False

    @property
    def axis(self):
        """The axis that the parameter corresponds to, as an integer indexed from zero.

        If :attr:`.Parameter.is_scalar` is true, then this will be None. Otherwise, it
        is an integer.
        """
        return self.__axis

    @property
    def cli_form(self):
        """The name of the parameter in a CLI argument format."""
        return f"-{self.name}"

    @property
    def index(self):
        """The index of the argument in the sequence of specified args."""
        return self.__index

    @property
    def was_used(self):
        """Whether this parameter was used in its associated :func:`.calculate` call."""
        return self._used


class TargetParameter(Parameter):
    """Cavity parameters which represent targets computed by cavcalc.

    .. note::

        You should never need to instantiate objects of this type manually. These
        are constructed during the calculation process.
    """

    def __init__(self, name_or_ptype: _Union[str, ParameterType], value: Q_ = None):
        super().__init__(name_or_ptype, value)



from .errors import CavCalcError


def quit_print(msg: str):
    """Prints `msg` to stdout and exits with code -1 if in CLI mode, otherwise
    raises a `CavCalcError` with the given `msg`."""
    from . import _SESSION

    if _SESSION.is_cli_mode:
        print("ERROR: " + msg)
        exit(-1)
    else:
        raise CavCalcError(msg)


def bug(msg: str):
    quit_print("Bug encountered! " + msg)



"""The argument groups loaded by the session argument parser."""

import argparse

from .. import Q_
from ..parameters import ArgParameter, ParameterType, valid_arguments, valid_targets
from ..__version__ import __version__

from ._types import float_file_range_t, mesh_t, limits_t, options_t


def make_arguments(parser: argparse.ArgumentParser):
    _make_target_arguments(parser)
    _make_physical_parameter_arguments(parser)
    _make_data_arguments(parser)
    _make_file_arguments(parser)
    _make_plotting_arguments(parser)
    _make_misc_arguments(parser)


def _make_target_arguments(parser: argparse.ArgumentParser):
    target_group = parser.add_argument_group(
        "Compute targets", "Arguments related to the target parameter to compute."
    )
    target_group.add_argument(
        "compute",
        choices=valid_targets + ("all",),
        help="Choice of parameter to compute. Default is 'all' for "
        "computing all available parameters from the given input values.",
        nargs="?",
        default="all",
    )
    target_group.add_argument(
        "-u",
        "--units",
        help="Units of output. Only used if a compute target was specified, i.e. it is ignored "
        "if computing all available targets. Note that if this option is specified then it "
        "will override the associated unit option in the config file.",
        type=str,
    )


def _make_physical_parameter_arguments(parser: argparse.ArgumentParser):
    physical_group = parser.add_argument_group(
        "Physical parameters",
        "Arguments which are the dependencies of the target parameter function. All of "
        "these arguments can be given as <value>[<units>] (single value, units optional), or "
        '"<start> <stop> <num> [<units>]" (range of values, units optional).',
    )

    for pname in valid_arguments:
        arg_p = ArgParameter(pname)
        physical_group.add_argument(
            arg_p.cli_form,
            help=arg_p.description,
            type=float_file_range_t,
            default=Q_("1064nm") if arg_p.ptype == ParameterType.WAVELENGTH else None,
        )


def _make_data_arguments(parser: argparse.ArgumentParser):
    data_group = parser.add_argument_group(
        "Data options", "Arguments for manipulating physical parameter ranges."
    )
    data_group.add_argument(
        "--mesh",
        dest="mesh",
        help="Construct mesh-grids from specified parameter combinations.",
        type=mesh_t,
    )


def _make_file_arguments(parser: argparse.ArgumentParser):
    file_group = parser.add_argument_group(
        "File options", "Arguments for loading and saving output instances."
    )
    file_group.add_argument(
        "-load",
        "--loadfile",
        help="Name of file from which to load a serialized cavcalc output object.",
        type=str,
    )
    file_group.add_argument(
        "-save",
        "--savefile",
        help="Name of file to which the output object will be saved.",
        type=str,
    )


def _make_plotting_arguments(parser: argparse.ArgumentParser):
    plotting_group = parser.add_argument_group(
        "Plotting options", "Arguments for generating plots."
    )
    plotting_group.add_argument(
        "--plot",
        dest="plot",
        action="store_true",
        help="Plot the target data on linear scales.",
    )
    plotting_group.add_argument(
        "--cmap",
        dest="cmap",
        help="Matplotlib compliant colormap to use for the generated image plot. Ignored "
        "if no plotting action was requested, or if no meshes were defined.",
        type=str,
    )
    plotting_group.add_argument(
        "--logxplot",
        dest="semilogx",
        action="store_true",
        help="Plot the target data with log-scale x-axis. Ignored if any meshes were defined.",
    )
    plotting_group.add_argument(
        "--logyplot",
        dest="semilogy",
        action="store_true",
        help="Plot the target data with log-scale y-axis. Ignored if any meshes were defined.",
    )
    plotting_group.add_argument(
        "--logplot",
        dest="loglog",
        action="store_true",
        help="Plot the target data on a log-log scale if no meshes were defined, otherwise "
        "otherwise uses a logarithmic normalisation scale for the image plot colorbar(s).",
    )
    plotting_group.add_argument(
        "--no-plot-display",
        dest="no_show_plot",
        action="store_true",
        help="Show the auto-generated plot(s).",
    )
    plotting_group.add_argument("--saveplot", help="Save the figure with the specified name.")
    plotting_group.add_argument(
        "--xlim",
        dest="xlim",
        help="Limits of x-axis.",
        type=limits_t,
    )
    plotting_group.add_argument(
        "--ylim",
        dest="ylim",
        help="Limits of y-axis.",
        type=limits_t,
    )
    plotting_group.add_argument(
        "--zlim",
        dest="zlim",
        help="Limits of color-bar scale. Ignored if no meshes were defined.",
        type=limits_t,
    )
    plotting_group.add_argument(
        "--plot-opts",
        dest="plot_opts",
        help="Optional keyword arguments to pass to relevant matplotlib plotting function.",
        type=options_t,
        default={},
    )
    parser.set_defaults(
        plot=False, semilogx=False, semilogy=False, loglog=False, no_show_plot=False
    )


def _make_misc_arguments(parser: argparse.ArgumentParser):
    misc_group = parser.add_argument_group(
        "Miscellaneous options",
        "Arguments for displaying information about cavcalc.",
    )
    misc_group.add_argument(
        "-V",
        "--version",
        help="Current version of cavcalc.",
        action="version",
        version=f"cavcalc v{__version__}",
    )



"""
The type and category enumerations for parameters.
"""

from enum import auto, Enum


class ParameterType(Enum):
    """An enumeration containing each of the physical parameters
    associated with an optical cavity."""

    BEAMSIZE = auto()
    BEAMSIZE_M1 = auto()
    BEAMSIZE_M2 = auto()
    CAV_GFACTOR = auto()
    CAV_LENGTH = auto()
    DIVERGENCE = auto()
    FINESSE = auto()
    FSR = auto()
    FWHM = auto()
    GFACTOR_M1 = auto()
    GFACTOR_M2 = auto()
    GFACTOR_SINGLE = auto()
    GOUY = auto()
    LOSS_M1 = auto()
    LOSS_M2 = auto()
    MODESEP = auto()
    POLE = auto()
    REFLECTIVITY_M1 = auto()
    REFLECTIVITY_M2 = auto()
    RES_ENHANCE_INTERNAL = auto()
    RES_ENHANCE_EXTERNAL = auto()
    ROC = auto()
    ROC_M1 = auto()
    ROC_M2 = auto()
    TRANSMISSION_INTENSITY_FRAC = auto()
    TRANSMISSION_M1 = auto()
    TRANSMISSION_M2 = auto()
    WAISTPOS = auto()
    WAISTSIZE = auto()
    WAVELENGTH = auto()


class ParameterCategory(Enum):
    """The categories for each physical parameter type."""

    Angle = auto()
    BeamRadius = auto()
    Curvature = auto()
    Distance = auto()
    Frequency = auto()
    Phase = auto()
    Power = auto()
    Stability = auto()
    Wave = auto()



"""Custom exception types defined by ``cavcalc``."""

# TODO (sjr) May want more granular exception types deriving
#            from this in the future
class CavCalcError(Exception):
    """The base error class for all ``cavcalc`` exceptions."""

    pass



import sys

from . import _SESSION
from ._main import main
from ._session import SessionType


def cli_entry():
    # Inform the back-end that we are now in CLI mode
    _SESSION.mode = SessionType.CLI

    # No argument specified, just default to help page
    if len(sys.argv) == 1:
        _SESSION.parser.print_help(sys.stderr)
        sys.exit(1)

    main()


if __name__ == "__main__":
    cli_entry()



import numpy as np
from typing import Any, Union
import warnings

from ._utils import _ReferencedQuantity
from .. import Q_
from .._exiters import bug, quit_print
from ..parameters import ArgParameter, ParameterType


class _Arguments:
    def __init__(self):
        self.__physical_args = {}

    @property
    def physical_args(self) -> dict[ParameterType, ArgParameter]:
        return self.__physical_args

    @property
    def is_all_target(self) -> bool:
        return self["compute"] == "all"

    @property
    def is_single_target(self) -> bool:
        return not self.is_all_target

    def __getitem__(self, argname: str) -> Union[ArgParameter, Any]:
        return getattr(self, argname)

    def has(self, argname: str) -> bool:
        """Checks that `argname` was specified as one of the arguments, and so
        is held by this namespace-like object."""
        return getattr(self, argname, None) is not None

    def process(self, arg_orders: dict[str, int]):
        self.__resolve_cross_refs()
        filter_phys_args = lambda: {name: v for name, v in vars(self).items() if isinstance(v, Q_)}
        phys_args = filter_phys_args()

        # Construct mesh-grids from these parameters in the order specified
        axis_orders = {}
        if self.mesh:
            if isinstance(self.mesh, bool):
                auto_mesh = tuple(
                    name for name in arg_orders.keys() if isinstance(phys_args[name].m, np.ndarray)
                )
                if not auto_mesh:
                    warnings.warn("Ignoring mesh argument, as no parameters are array-like.")

                self.mesh = (auto_mesh,)

            for param_combo in self.mesh:
                if len(param_combo) == 1:
                    pname = param_combo[0]
                    axis_orders[pname] = 0
                    warnings.warn(f"Ignoring single parameter '{pname}' in mesh.")
                    continue

                try:
                    arg_ps = tuple(phys_args[pname] for pname in param_combo)
                except KeyError as ex:
                    quit_print(f"Mesh parameter {str(ex)} was not specified!")

                if any(not isinstance(p.m, np.ndarray) for p in arg_ps):
                    quit_print(
                        "One or more of the specified mesh parameters is not array-like! Cannot "
                        "construct a mesh-grid from scalar arguments."
                    )
                meshes = np.meshgrid(*(arg_p.m for arg_p in arg_ps), indexing="ij")
                for i, (pname, mesh) in enumerate(zip(param_combo, meshes)):
                    setattr(self, pname, Q_(mesh, phys_args[pname].units))
                    axis_orders[pname] = i

            # For the remaining array-like args, which were not in any mesh combination, just
            # set the axis number to the *current* number of entries in axis_orders
            n_axes = len(axis_orders)
            for pname, v in phys_args.items():
                if pname not in axis_orders and isinstance(v.m, np.ndarray):
                    axis_orders[pname] = n_axes

            # Re-construct physical args dict, as we've now modified attributes
            phys_args = filter_phys_args()
        else:
            for pname, v in phys_args.items():
                if isinstance(v.m, np.ndarray):
                    axis_orders[pname] = 0

        # Make Parameter instances for each physical argument given
        for name, q in phys_args.items():
            param_obj = ArgParameter(
                name, q, index=arg_orders.get(name), axis=axis_orders.get(name)
            )
            setattr(self, name, param_obj)
            self.__physical_args[param_obj.ptype] = param_obj

        for m in (1, 2):
            # Set the loss arguments to zero if either of R<m>, T<m> given but L<m> not
            if (self.has(f"R{m}") or self.has(f"T{m}")) and not self.has(f"L{m}"):
                Lm = ArgParameter(f"L{m}", Q_(0))
                setattr(self, f"L{m}", Lm)
                self.__physical_args[Lm.ptype] = Lm

        self.__verify()

    def __resolve_cross_refs(self):
        for name, v in vars(self).copy().items():
            if isinstance(v, _ReferencedQuantity):
                if name == v.ref_name:
                    quit_print(f"Argument '{name}' references itself!")

                q = getattr(self, v.ref_name)
                ref_chain = set([name, v.ref_name])
                while isinstance(q, _ReferencedQuantity):
                    if q.ref_name in ref_chain:
                        quit_print(
                            "Encountered a reference loop whilst trying to "
                            f"resolve the parameter '{name}'."
                        )
                    q = getattr(self, q.ref_name)

                if q is None:
                    quit_print(f"Argument '{name}' references a parameter which was not given.")
                elif not isinstance(q, Q_):
                    bug(f"Argument '{name}' references a parameter which is not a quantity.")

                setattr(self, name, q)

    def __verify(self):
        """Checks that the arguments held by this namespace-like object make sense
        and are consistent with one another."""
        if self.has("units"):
            if self.is_all_target:
                warnings.warn(
                    "Specifying output units is only supported in single target mode. Use "
                    "a 'cavcalc.ini' config file to override units in multi target mode, or "
                    "interact with cavcalc via the Python API for tasks requiring more "
                    "customisation."
                )

        if self.has("loadfile") and self.physical_args:
            # We will always have wavelength in the physical_args as this takes on a default value
            if tuple(self.physical_args.keys()) != (ParameterType.WAVELENGTH,):
                quit_print(
                    "You cannot give physical parameter arguments when loading a "
                    "serialized cavcalc output object."
                )

        self.__verify_physical_parameters()

    def __verify_physical_parameters(self):
        if self.has("T1") and self.has("R1"):
            quit_print(
                "Incorrect usage! Cannot specify both reflectivity and transmission of first mirror."
            )
        if self.has("T2") and self.has("R2"):
            quit_print(
                "Incorrect usage! Cannot specify both reflectivity and transmission of second mirror."
            )

        # Check {R, T, L} are within [0, 1] for both mirrors
        for p in "R", "T", "L":
            for m in (1, 2):
                if self.has(f"{p}{m}"):
                    v = self[f"{p}{m}"].value.m
                    if np.any((v < 0) | (v > 1)):
                        quit_print(f"{p}{m} is invalid. Value(s) must satisfy 0 <= {p} <= 1.")

        # And, if necessary, check {(R + L) | (T + L)} < 1 for both mirrors too
        for p in "R", "T":
            for m in (1, 2):
                if (Pm := self[f"{p}{m}"]) and (Lm := self[f"L{m}"]):
                    Pv = Pm.value.m
                    Lv = Lm.value.m
                    try:
                        if np.any((Pv + Lv) > 1):
                            quit_print(
                                f"Invalid combination of {p}{m} and L{m} given. Value(s) "
                                f"must satisfy {p} + L <= 1."
                            )
                    except ValueError:
                        quit_print(
                            f"Encountered inconsistent array shapes for {p}{m} and L{m}, with "
                            f"shapes: {Pv.shape}, {Lv.shape}, respectively. Could not broadcast "
                            "these arrays together."
                        )

        if L := self["L"]:
            if np.any(L.value.m < 0):
                quit_print("L is invalid. Cavity lengths must be non-negative.")

        if wl := self["wl"]:
            if np.any(wl.value.m <= 0):
                quit_print("wl is invalid. Wavelengths must be positive.")

        # Make sure no illogical combinations of stability were given...
        has_g = self.has("g")
        has_g1 = self.has("g1")
        has_g2 = self.has("g2")
        has_gs = self.has("gs")
        if has_g and (has_g1 or has_g2 or has_gs):
            instruct = (
                "To specify a non-symmetric cavity please use both 'g1' and 'g2', do not give a "
                "value for 'g'. Or, to specify a symmetric cavity please use either 'g' or 'gs'; "
                "g is then gs * gs where gs = g1 = g2."
            )

            if has_g1 or has_g2:
                msg = (
                    "Incorrect usage! Cavity g-factor and individual mirror g-factors "
                    "cannot be specified simultaneously."
                )
            elif has_gs:
                msg = (
                    "Incorrect usage! Cavity g-factor and symmetric mirror g-factors "
                    "cannot be specified simultaneously."
                )

            quit_print(msg + "\n\n" + instruct)

        if has_gs and (has_g1 or has_g2):
            quit_print(
                "Incorrect usage! Individual mirror g-factors and symmetric "
                "mirror g-factor cannot be specified simultaneously."
            )

        # ... and same goes for combinations of RoCs...
        if self.has("Rc") and (self.has("Rc1") or self.has("Rc2")):
            quit_print(
                "Incorrect usage! Symmetric curvature, 'Rc', and individual (non-symmetric), 'Rc1' and 'Rc2', "
                "curvatures cannot be specified simultaneously.\n\nTo specify a non-symmetric cavity "
                "please use both 'Rc1' and 'Rc2', do not give a value for 'Rc'. To specify a symmetric "
                "cavity please use just 'Rc'."
            )

        # ... and beam-sizes
        if self.has("w") and (self.has("w1") or self.has("w2")):
            quit_print(
                "Incorrect usage! Symmetric beam-size, 'w', and individual (non-symmetric), 'w1' and 'w2', "
                "beam-sizes cannot be specified simultaneously.\n\nTo specify a non-symmetric cavity "
                "please use both 'w1' and 'w2', do not give a value for 'w'. To specify a symmetric "
                "cavity please use just 'w'."
            )
        for p in "w", "w1", "w2":
            if P := self[p]:
                if np.any(P.value.m < 0):
                    quit_print(f"{p} is invalid. Beam sizes must be non-negative.")



"""
A package for computing Fabry-Perot optical cavity parameters.

The main port-of-call, for interacting with cavcalc, should be via the
recommended "single-function interface" :func:`.calculate`. See
:ref:`module` for in-depth details on using cavcalc programmatically.
"""

import configparser as _configparser
import os as _os
import shutil as _shutil
import warnings as _warnings

try:
    from .__version__ import __version__
except ImportError:
    __version__ = "?.?.?"


import pint as _pint


ureg = _pint.UnitRegistry(cache_folder=":auto:")
_pint.set_application_registry(ureg)  # ensure compatibility when pickling / unpickling data
Q_ = ureg.Quantity

# Construct the session singleton, holding environment information
from ._session import _CavCalcSession

_SESSION = _CavCalcSession()


_HERE = _os.path.dirname(_os.path.realpath(__file__))


def _get_usr_config_dir():
    return _os.path.join(
        _os.environ.get("APPDATA")
        or _os.environ.get("XDG_CONFIG_HOME", _os.path.expanduser("~/.config")),
        "cavcalc",
    )


def _write_config_file():
    from ._exiters import quit_print as _quit_print

    usr_conf_path = _get_usr_config_dir()
    _os.makedirs(usr_conf_path, exist_ok=True)

    package_conf_file = _os.path.join(_HERE, "cavcalc.ini")
    if not _os.path.exists(package_conf_file):
        _quit_print(
            f"No 'cavcalc.ini' file present in the package install directory: {_HERE}. "
            "\n\nTry re-installing cavcalc, and file a bug report if this does not work."
        )

    usr_conf_file = _os.path.join(usr_conf_path, "cavcalc.ini")
    if not _os.path.isfile(usr_conf_file):
        _shutil.copyfile(package_conf_file, usr_conf_file)


def _read_config_file():
    from ._exiters import quit_print as _quit_print

    config = _configparser.ConfigParser()

    # Read in order (if they exist):
    #    1. package config file
    #    2. user .config directory file
    #    3. current directory file
    # with options from later successfully read files taking priority.
    read_files = config.read(
        tuple(_os.path.join(d, "cavcalc.ini") for d in (_HERE, _get_usr_config_dir(), ""))
    )
    if not read_files:
        _quit_print("Unable to read any 'cavcalc.ini' config files!")

    return config


def _process_config(config: _configparser.ConfigParser):
    from ._exiters import quit_print as _quit_print

    if "plotting" not in config:
        _warnings.warn(
            "No [plotting] section in any loaded cavcalc config file. This may lead to errors."
        )

    if "units" not in config:
        _quit_print(f"No [units] section present in any loaded cavcalc config file.")

    parents_map = {
        "div": "Angles",
        "w": "Beamsizes",
        "w1": "Beamsizes",
        "w2": "Beamsizes",
        "w0": "Beamsizes",
        "Rc": "RoCs",
        "Rc1": "RoCs",
        "Rc2": "RoCs",
        "L": "Distances",
        "z0": "Distances",
        "FSR": "Frequencies",
        "FWHM": "Frequencies",
        "modesep": "Frequencies",
        "pole": "Frequencies",
        "gouy": "Phases",
        "wl": "Waves",
    }

    units_sec = config["units"]

    # Ensure each parameter, which does *not* have a units override
    # specified, inherits units from its corresponding parent category
    for param, parent in parents_map.items():
        if not units_sec.get(param):
            parent_units = units_sec.get(parent)
            if not parent_units:
                _quit_print(
                    f"No units specified in category '{parent.capitalize()}' for "
                    "any loaded cavcalc config file."
                )

            units_sec[param] = parent_units


_write_config_file()
_CONFIG = _read_config_file()
_process_config(_CONFIG)


from .calculate import calculate
from .env import configure
from .output import load
from .parameters import ParameterType, get_default_units



"""
The underlying functions used by cavcalc for computing the optical
resonator properties.

.. note::

    Whilst the underlying functions in this sub-package are publicly accessible, it is
    recommended that you instead use the single-function interface :func:`.calculate` for
    computing any of the cavity properties that you need; as that will result in a nice output
    object for accessing results.

    If, for whatever reason, you do choose to use any of the functions listed in the
    sub-modules below, please note that all of these *require* that the arguments passed to
    them are ``pint.Quantity`` instances.
"""

from . import resonance
from . import symmetric
from . import asymmetric



"""Local environment settings modification for cavcalc."""

import os
import warnings

from . import _CONFIG
from .errors import CavCalcError
from .parameters import ParameterCategory
from .parameters.tools import get_names


class _ConfigureContext:
    def __init__(self, previous_units: dict[str, str], original_rc_params: dict):
        self.__prev = previous_units
        self.__orig_rc = original_rc_params

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        _CONFIG["units"].update(self.__prev)
        if self.__orig_rc:
            import matplotlib.pyplot as plt

            dict.update(plt.rcParams, self.__orig_rc)


def configure(plt_style=None, **param_units):
    """Configure cavcalc plotting style, and parameter unit overrides.

    This function can be used both regularly, and via context-managed scopes using
    the ``with`` statement. In the latter case, any changes made via these configuration
    options will be reset on exit from the context-managed block.

    Parameters
    ----------
    plt_style : str | dict | Path | list
        The style-sheet to use, see :func:`matplotlib.pyplot.style.use` for details. Note
        that, in addition to the options specified in the linked matplotlib documentation,
        one can set this to ``"cavcalc"`` to use the default style-sheet that this package
        provides. This argument defaults to ``None``, such that no style modification
        is performed (and matplotlib not imported); this way, the default behaviour of
        this function is to have as minimal impact on the users' development workflow
        as possible.

    **param_units : Keyword Arguments
        Keyword arguments specifying overrides for the units of any cavcalc parameter or
        parameter category. The names of these correspond exactly to those in the units
        section of a ``cavcalc.ini`` config file.

    Examples
    --------
    If you simply want to set-up the plotting style to be the same as the default
    style used when running cavcalc from the command line, then do::

        import cavcalc as cc
        cc.configure(plt_style="cavcalc")

    or, if you only want to use this style-sheet temporarily::

        import numpy as np

        with cc.configure(plt_style="cavcalc"):
            # do stuff with cavcalc, e.g:
            cc.calculate("w", L=1, Rc=np.linspace(1, 5, 101)).plot()
    """
    if plt_style:
        import matplotlib.pyplot as plt

        if isinstance(plt_style, str) and plt_style.casefold() == "cavcalc":
            here, _ = os.path.split(os.path.realpath(__file__))
            cc_style_file = os.path.join(here, "_default.mplstyle")
            if not os.path.isfile(cc_style_file):
                raise CavCalcError("Could not locate cavcalc package style-sheet file!")

            plt_style = cc_style_file

        orig_rc = dict(plt.rcParams.copy())
        plt.style.use(plt_style)
    else:
        orig_rc = None

    children_map = {
        "angles": ("div",),
        "beamsizes": ("w", "w1", "w2", "w0"),
        "rocs": ("Rc", "Rc1", "Rc2"),
        "distances": ("L", "z0"),
        "frequencies": ("FSR", "FWHM", "modesep", "pole"),
        "phases": ("gouy",),
        "waves": ("wl",),
    }
    for pcat, params in children_map.items():
        if pcat_units := param_units.get(pcat):
            for p in params:
                # Only use category override if specific units
                # override is *not* also in the kwargs
                param_units.setdefault(p, pcat_units)

            del param_units[pcat]

    dimensionless_params = get_names(ParameterCategory.Power, ParameterCategory.Stability)

    cfg_units = _CONFIG["units"]
    prev = {}
    for p, units in param_units.items():
        if curr_units := cfg_units.get(p):
            prev[p] = curr_units
            cfg_units[p] = units
        else:
            msg = "Ignoring units override for "
            if p in dimensionless_params:
                warnings.warn(msg + f"dimensionless parameter: {p}")
            else:
                warnings.warn(msg + f"invalid parameter name: {p}")

    return _ConfigureContext(prev, orig_rc)
